# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import datetime
import hashlib
import hmac
import json
import uuid
import warnings

import pytz
import requests
from six.moves.urllib import parse as urlparse

from .common import (
    CANCELED,
    ERROR,
    FORM,
    PAID,
    Form,
    PaymentCommon,
    PaymentResponse,
    ResponseError,
    _,
    force_text,
)

__all__ = ['Payment']


class Payment(PaymentCommon):
    """
    Payment backend module for the ATOS/SIPS system used by many French banks.

    The necessary options are:
    - merchant_id
    - secret_key
    - normal_return_url
    - automatic_return_url

    It was writtent using the documentation from file:

        Worldline Benelux_Sips_Technical_integration_guide_Version_1.5.pdf

    """

    has_free_transaction_id = True

    URL = {
        'test': 'https://payment-webinit.simu.sips-services.com/paymentInit',
        'prod': 'https://payment-webinit.sips-services.com/paymentInit',
    }
    WS_URL = {
        'test': 'https://office-server.test.sips-services.com',
        'prod': 'https://office-server.sips-services.com',
    }

    INTERFACE_VERSION = 'HP_2.3'
    RESPONSE_CODES = {
        '00': 'Authorisation accepted',
        '02': 'Authorisation request to be performed via telephone with the issuer, as the '
        'card authorisation threshold has been exceeded, if the forcing is authorised for '
        'the merchant',
        '03': 'Invalid distance selling contract',
        '05': 'Authorisation refused',
        '12': 'Invalid transaction, verify the parameters transferred in the request.',
        '14': 'Invalid bank details or card security code',
        '17': 'Buyer cancellation',
        '24': 'Operation impossible. The operation the merchant wishes to perform is not '
        'compatible with the status of the transaction.',
        '25': 'Transaction not found in the Sips database',
        '30': 'Format error',
        '34': 'Suspicion of fraud',
        '40': 'Function not supported: the operation that the merchant would like to perform '
        'is not part of the list of operations for which the merchant is authorised',
        '51': 'Amount too high',
        '54': 'Card is past expiry date',
        '60': 'Transaction pending',
        '63': 'Security rules not observed, transaction stopped',
        '75': 'Number of attempts at entering the card number exceeded',
        '90': 'Service temporarily unavailable',
        '94': 'Duplicated transaction: for a given day, the TransactionReference has already been ' 'used',
        '97': 'Timeframe exceeded, transaction refused',
        '99': 'Temporary problem at the Sips Office Server level',
    }
    TEST_MERCHANT_ID = '002001000000001'

    description = {
        'caption': 'SIPS 2',
        'parameters': [
            {
                'name': 'platform',
                'caption': _('Platform'),
                'default': 'test',
                'choices': ['test', 'prod'],
                'required': True,
            },
            {
                'name': 'merchant_id',
                'caption': _('Merchant ID'),
                'default': TEST_MERCHANT_ID,
                'required': True,
            },
            {
                'name': 'secret_key',
                'caption': _('Secret Key'),
                'default': '002001000000001_KEY1',
                'required': True,
            },
            {
                'name': 'key_version',
                'caption': _('Key Version'),
                'default': '1',
                'required': True,
            },
            {
                'name': 'normal_return_url',
                'caption': _('Normal return URL'),
                'default': '',
                'required': True,
            },
            {
                'name': 'automatic_return_url',
                'caption': _('Automatic return URL'),
                'required': False,
            },
            {
                'name': 'currency_code',
                'caption': _('Currency code'),
                'default': '978',
                'choices': ['978'],
                'required': True,
            },
            {
                'name': 'capture_mode',
                'caption': _('Capture Mode'),
                'default': 'AUTHOR_CAPTURE',
                'choices': ['AUTHOR_CAPTURE', 'IMMEDIATE', 'VALIDATION'],
                'required': True,
            },
            {
                'name': 'capture_day',
                'caption': _('Capture Day'),
                'required': False,
            },
            {'name': 'payment_means', 'caption': _('Payment Means'), 'required': False},
            {
                'name': 'timezone',
                'caption': _('SIPS server Timezone'),
                'default': 'Europe/Paris',
                'required': False,
            },
        ],
    }

    def encode_data(self, data):
        return '|'.join('%s=%s' % (force_text(key), force_text(value)) for key, value in data.items())

    def seal_data(self, data):
        s = self.encode_data(data)
        s += force_text(self.secret_key)
        s = s.encode('utf-8')
        s = hashlib.sha256(s).hexdigest()
        return s

    def get_data(self):
        data = collections.OrderedDict()
        data['merchantId'] = self.merchant_id
        data['keyVersion'] = self.key_version
        data['normalReturnUrl'] = self.normal_return_url
        if self.automatic_return_url:
            data['automaticResponseUrl'] = self.automatic_return_url
        data['currencyCode'] = self.currency_code
        data['captureMode'] = self.capture_mode
        if self.capture_day:
            data['captureDay'] = self.capture_day
        if self.payment_means:
            data['paymentMeanBrandList'] = self.payment_means
        return data

    def get_url(self):
        return self.URL[self.platform]

    def request(
        self,
        amount,
        name=None,
        first_name=None,
        last_name=None,
        address=None,
        email=None,
        phone=None,
        orderid=None,
        info1=None,
        info2=None,
        info3=None,
        next_url=None,
        transaction_id=None,
        **kwargs,
    ):
        data = self.get_data()
        # documentation:
        # https://documentation.sips.worldline.com/fr/WLSIPS.801-MG-Presentation-generale-de-la-migration-vers-Sips-2.0.html#ariaid-title20
        transactionReference = transaction_id or uuid.uuid4().hex
        data['transactionReference'] = transactionReference
        data['orderId'] = orderid or transactionReference
        if info1:
            data['statementReference'] = force_text(info1)
        else:
            data['statementReference'] = data['transactionReference']
        data['amount'] = self.clean_amount(amount)
        if email:
            data['billingContact.email'] = email
        if 'capture_day' in kwargs:
            data['captureDay'] = kwargs.get('capture_day')
        normal_return_url = self.normal_return_url
        if next_url and not normal_return_url:
            warnings.warn(
                "passing next_url to request() is deprecated, " "set normal_return_url in options",
                DeprecationWarning,
            )
            normal_return_url = next_url
        if normal_return_url:
            data['normalReturnUrl'] = normal_return_url
        form = Form(
            url=self.get_url(),
            method='POST',
            fields=[
                {'type': 'hidden', 'name': 'Data', 'value': self.encode_data(data)},
                {
                    'type': 'hidden',
                    'name': 'Seal',
                    'value': self.seal_data(data),
                },
                {
                    'type': 'hidden',
                    'name': 'InterfaceVersion',
                    'value': self.INTERFACE_VERSION,
                },
            ],
        )
        self.logger.debug('emitting request %r', data)
        return transactionReference, FORM, form

    @classmethod
    def decode_data(cls, data):
        data = data.split('|')
        data = [map(force_text, p.split('=', 1)) for p in data]
        return collections.OrderedDict(data)

    def check_seal(self, data, seal):
        return seal == self.seal_data(data)

    response_code_to_result = {
        '00': PAID,
        '17': CANCELED,
    }

    def response(self, query_string, **kwargs):
        form = urlparse.parse_qs(query_string)
        if not set(form) >= {'Data', 'Seal', 'InterfaceVersion'}:
            raise ResponseError('missing Data, Seal or InterfaceVersion')
        self.logger.debug('received query string %r', form)
        data = self.decode_data(form['Data'][0])
        seal = form['Seal'][0]
        self.logger.debug('parsed response %r seal %r', data, seal)
        signed = self.check_seal(data, seal)
        response_code = data['responseCode']
        transaction_id = data.get('transactionReference')
        result = self.response_code_to_result.get(response_code, ERROR)
        merchant_id = data.get('merchantId')
        test = merchant_id == self.TEST_MERCHANT_ID
        transaction_date = None
        if 'transactionDateTime' in data:
            try:
                transaction_date = datetime.datetime.strptime(
                    data['transactionDateTime'], '%Y-%m-%d %H:%M:%S'
                )
            except (ValueError, TypeError):
                pass
            else:
                sips_tz = pytz.timezone(self.timezone)
                transaction_date = sips_tz.localize(transaction_date)
        return PaymentResponse(
            result=result,
            signed=signed,
            bank_data=data,
            order_id=transaction_id,
            transaction_id=data.get('authorisationId'),
            bank_status=self.RESPONSE_CODES.get(response_code, 'unknown code - ' + response_code),
            test=test,
            transaction_date=transaction_date,
        )

    def get_seal_for_json_ws_data(self, data):
        data_to_send = []
        for key in sorted(data.keys()):
            if key in ('keyVersion', 'sealAlgorithm', 'seal'):
                continue
            data_to_send.append(force_text(data[key]))
        data_to_send_str = ''.join(data_to_send).encode('utf-8')
        return hmac.new(
            force_text(self.secret_key).encode('utf-8'), data_to_send_str, hashlib.sha256
        ).hexdigest()

    def perform_cash_management_operation(self, endpoint, data):
        data['merchantId'] = self.merchant_id
        data['interfaceVersion'] = 'CR_WS_2.3'
        data['keyVersion'] = self.key_version
        data['currencyCode'] = self.currency_code
        data['seal'] = self.get_seal_for_json_ws_data(data)
        url = self.WS_URL.get(self.platform) + '/rs-services/v2/cashManagement/%s' % endpoint
        self.logger.debug('posting %r to %s endpoint', data, endpoint)
        response = requests.post(
            url,
            data=json.dumps(data),
            headers={'content-type': 'application/json', 'accept': 'application/json'},
        )
        self.logger.debug('received %r', response.content)
        response.raise_for_status()
        json_response = response.json()
        if self.platform == 'prod':
            # test environment doesn't set seal
            if json_response.get('seal') != self.get_seal_for_json_ws_data(json_response):
                raise ResponseError('wrong signature on response')
        if json_response.get('responseCode') != '00':
            raise ResponseError('non-zero response code (%s)' % json_response)
        return json_response

    def cancel(self, amount, bank_data, **kwargs):
        data = {}
        data['operationAmount'] = self.clean_amount(amount)
        data['transactionReference'] = bank_data.get('transactionReference')
        return self.perform_cash_management_operation('cancel', data)

    def validate(self, amount, bank_data, **kwargs):
        data = {}
        data['operationAmount'] = self.clean_amount(amount)
        data['transactionReference'] = bank_data.get('transactionReference')
        return self.perform_cash_management_operation('validate', data)

    def diagnostic(self, amount, bank_data, **kwargs):
        data = {}
        data['transactionReference'] = bank_data.get('transactionReference')
        data['merchantId'] = self.merchant_id
        data['interfaceVersion'] = 'DR_WS_2.9'
        data['keyVersion'] = self.key_version
        data['seal'] = self.get_seal_for_json_ws_data(data)
        url = self.WS_URL.get(self.platform) + '/rs-services/v2/diagnostic/getTransactionData'
        self.logger.debug('posting %r to %s endpoint', data, 'diagnostic')
        response = requests.post(
            url,
            data=json.dumps(data),
            headers={
                'content-type': 'application/json',
                'accept': 'application/json',
            },
        )
        self.logger.debug('received %r', response.content)
        response.raise_for_status()
        return response.json()

    @classmethod
    def guess(self, *, method=None, query_string=None, body=None, headers=None, backends=(), **kwargs):
        for content in [query_string, body]:
            if isinstance(content, bytes):
                try:
                    content = content.decode()
                except UnicodeDecodeError:
                    pass
            if isinstance(content, str):
                fields = urlparse.parse_qs(content)
                if not set(fields) >= {'Data', 'Seal', 'InterfaceVersion'}:
                    continue
                data = self.decode_data(fields['Data'][0])
                if 'transactionReference' in data:
                    return data['transactionReference']
        return None



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import httmock
import lxml.etree as ET
import pytest
from requests import Session
from requests.adapters import HTTPAdapter


def pytest_addoption(parser):
    parser.addoption("--save-http-session", action="store_true", help="save HTTP session")
    parser.addoption("--target-url", help="target URL")


class LoggingAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.history = []
        super().__init__(*args, **kwargs)

    def send(self, request, *args, **kwargs):
        response = super().send(request, *args, **kwargs)
        self.history.append((request, response))
        return response


def xmlindent(content):
    if hasattr(content, 'encode') or hasattr(content, 'decode'):
        content = ET.fromstring(content)
    return ET.tostring(content, pretty_print=True).decode('utf-8', 'ignore')


@pytest.fixture
def record_http_session(request):
    module_name = request.module.__name__.split('test_', 1)[-1]
    function_name = request.function.__name__
    save = request.config.getoption('--save-http-session')
    filename = 'tests/data/%s-%s.json' % (module_name, function_name)

    def is_xml_content_type(r):
        headers = r.headers
        content_type = headers.get('content-type')
        return content_type and content_type.startswith(('text/xml', 'application/xml'))

    if save:
        session = Session()
        adapter = LoggingAdapter()
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        try:
            yield session
        finally:
            with open(filename, 'w') as fd:
                history = []

                for request, response in adapter.history:
                    request_content = request.body or b''
                    response_content = response.content or b''

                    if is_xml_content_type(request):
                        request_content = xmlindent(request_content)
                    else:
                        request_content = request_content.decode('utf-8')
                    if is_xml_content_type(response):
                        response_content = xmlindent(response_content)
                    else:
                        response_content = response_content.decode('utf-8')
                    history.append((request_content, response_content))
                json.dump(history, fd)
    else:
        with open(filename) as fd:
            history = json.load(fd)

        class Mocker:
            counter = 0

            @httmock.urlmatch()
            def mock(self, url, request):
                expected_request_content, response_content = history[self.counter]
                self.counter += 1
                if expected_request_content:
                    request_content = request.body or b''
                    if is_xml_content_type(request):
                        request_content = xmlindent(request_content)
                    else:
                        request_content = request_content.decode('utf-8')
                    assert request_content == expected_request_content
                return response_content

        with httmock.HTTMock(Mocker().mock):
            yield None


@pytest.fixture
def target_url(request):
    return request.config.getoption('--target-url') or 'https://target.url/'



#
# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import json

import pytest
import zeep.transports

import eopayment


@pytest.fixture
def saga(record_http_session):
    if record_http_session:
        from eopayment import saga

        saga._zeep_transport = zeep.transports.Transport(session=record_http_session)
        try:
            yield None
        finally:
            saga._zeep_transport = None
    else:
        yield


@pytest.fixture
def backend_factory(saga, target_url):
    def factory(**kwargs):
        return eopayment.Payment(
            'saga',
            dict(
                {
                    'base_url': target_url,
                    'num_service': '868',
                    'compte': '70688',
                    'automatic_return_url': 'https://automatic.notif.url/automatic/',
                    'normal_return_url': 'https://normal.notif.url/normal/',
                },
                **kwargs,
            ),
        )

    return factory


def test_error_parametrage(backend_factory):
    payment = backend_factory(num_service='1', compte='1')
    with pytest.raises(eopayment.PaymentException, match='Impossible de déterminer le paramétrage'):
        transaction_id, kind, url = payment.request(
            amount='10.00', email='john.doe@example.com', subject='Réservation concert XYZ numéro 1234'
        )


def test_request(backend_factory):
    transaction_id, kind, url = backend_factory().request(
        amount='10.00', email='john.doe@example.com', subject='Réservation concert XYZ numéro 1234'
    )
    assert transaction_id == '347b2060-1a37-11eb-af92-0213ad91a103'
    assert kind == eopayment.URL
    assert (
        url == 'https://www.tipi.budget.gouv.fr/tpa/paiementws.web?idop=347b2060-1a37-11eb-af92-0213ad91a103'
    )


def test_response(backend_factory):
    response = backend_factory().response('idop=28b52f40-1ace-11eb-8ce3-0213ad91a104', redirect=False)
    assert response.__dict__ == {
        'bank_data': {
            'email': 'john.doe@entrouvert.com',
            'etat': 'paye',
            'id_tiers': '-1',
            'montant': '10.00',
            'num_service': '868',
            'numcp': '70688',
            'numcpt_lib_ecriture': 'COUCOU',
        },
        'bank_status': 'paid',
        'order_id': '28b52f40-1ace-11eb-8ce3-0213ad91a104',
        'result': 3,
        'return_content': None,
        'signed': True,
        'test': False,
        'transaction_date': None,
        'transaction_id': '28b52f40-1ace-11eb-8ce3-0213ad91a104',
    }
    # Check bank_data is JSON serializable
    json.dumps(response.bank_data)



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from xml.etree import ElementTree as ET

import pytest
from six.moves.urllib import parse as urllib

import eopayment
import eopayment.ogone as ogone
from eopayment import ResponseError

PSPID = '2352566ö'


@pytest.fixture(params=[None, 'iso-8859-1', 'utf-8'])
def params(request):
    params = {
        'environment': ogone.ENVIRONMENT_TEST,
        'pspid': PSPID,
        'sha_in': 'sécret',
        'sha_out': 'sécret',
        'automatic_return_url': 'http://example.com/autömatic_réturn_url',
    }
    encoding = request.param
    if encoding:
        params['encoding'] = encoding
    return params


def test_request(params):
    ogone_backend = eopayment.Payment('ogone', params)
    amount = '42.42'
    order_id = 'my ordér'
    reference, kind, what = ogone_backend.request(amount=amount, orderid=order_id, email='foo@example.com')
    assert len(reference) == 32
    root = ET.fromstring(str(what))
    assert root.tag == 'form'
    assert root.attrib['method'] == 'POST'
    assert root.attrib['action'] == ogone.ENVIRONMENT_TEST_URL
    values = {
        'CURRENCY': 'EUR',
        'ORDERID': order_id,
        'PSPID': PSPID,
        'EMAIL': 'foo@example.com',
        'AMOUNT': amount.replace('.', ''),
        'LANGUAGE': 'fr_FR',
        'COMPLUS': reference,
    }
    values.update({'SHASIGN': ogone_backend.backend.sha_sign_in(values)})
    for node in root:
        assert node.attrib['type'] in ('hidden', 'submit')
        assert set(node.attrib.keys()), {'type', 'name' == 'value'}
        name = node.attrib['name']
        if node.attrib['type'] == 'hidden':
            assert name in values
            assert node.attrib['value'] == values[name]


def test_response(params):
    ogone_backend = eopayment.Payment('ogone', params)
    order_id = 'myorder'
    data = {
        'orderid': 'myorder',
        'status': '9',
        'payid': '3011229363',
        'cn': 'Usér',
        'ncerror': '0',
        'trxdate': '10/24/16',
        'acceptance': 'test123',
        'currency': 'eur',
        'amount': '7.5',
    }
    data['shasign'] = ogone_backend.backend.sha_sign_out(data, encoding=params.get('encoding', 'iso-8859-1'))
    # uniformize to utf-8 first
    for k in data:
        data[k] = eopayment.common.force_byte(data[k], encoding=params.get('encoding', 'iso-8859-1'))
    response = ogone_backend.response(urllib.urlencode(data))
    assert response.signed
    assert response.order_id == order_id


def test_iso_8859_1_response():
    params = {
        'environment': ogone.ENVIRONMENT_TEST,
        'pspid': PSPID,
        'sha_in': 'sécret',
        'sha_out': 'sécret',
        'automatic_return_url': 'http://example.com/autömatic_réturn_url',
    }
    ogone_backend = eopayment.Payment('ogone', params)
    order_id = 'lRXK4Rl1N2yIR3R5z7Kc'
    backend_response = (
        'orderID=lRXK4Rl1N2yIR3R5z7Kc&currency=EUR&amount=7%2E5'
        '&PM=CreditCard&ACCEPTANCE=test123&STATUS=9'
        '&CARDNO=XXXXXXXXXXXX9999&ED=0118'
        '&CN=Miha%EF+Serghe%EF&TRXDATE=10%2F24%2F16'
        '&PAYID=3011228911&NCERROR=0&BRAND=MasterCard'
        '&IP=80%2E12%2E92%2E47&SHASIGN=C429BE892FACFBFCE5E2CC809B102D866DD3D48C'
    )
    response = ogone_backend.response(backend_response)
    assert response.signed
    assert response.order_id == order_id


def test_bad_response(params):
    ogone_backend = eopayment.Payment('ogone', params)
    data = {'payid': '32100123', 'status': 9, 'ncerror': 0}
    with pytest.raises(ResponseError, match='missing ORDERID, PAYID, STATUS or NCERROR'):
        ogone_backend.response(urllib.urlencode(data))


def test_bank_transfer_response(params):
    ogone_backend = eopayment.Payment('ogone', params)
    data = {
        'orderid': 'myorder',
        'status': '41',
        'payid': '3011229363',
        'cn': 'User',
        'ncerror': '0',
        'trxdate': '10/24/16',
        'brand': 'Bank transfer',
        'pm': 'bank transfer',
        'currency': 'eur',
        'amount': '7.5',
        'shasign': '944CBD1E010BA4945415AE4B16CC40FD533F6CE2',
    }
    # uniformize to expected encoding
    for k in data:
        data[k] = eopayment.common.force_byte(data[k], encoding=params.get('encoding', 'iso-8859-1'))
    response = ogone_backend.response(urllib.urlencode(data))
    assert response.signed
    assert response.result == eopayment.WAITING

    # check utf-8 based signature is also ok
    data['shasign'] = b'0E35F687ACBEAA6CA769E0ADDBD0863EB6C1678A'
    response = ogone_backend.response(urllib.urlencode(data))
    assert response.signed
    assert response.result == eopayment.WAITING

    # check invalid signature is not marked ok
    data['shasign'] = b'0000000000000000000000000000000000000000'
    response = ogone_backend.response(urllib.urlencode(data))
    assert not response.signed



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

import pytest
from six.moves.urllib.parse import parse_qs, urlparse

import eopayment


@pytest.fixture
def backend():
    options = {
        'automatic_notification_url': 'http://example.com/direct_notification_url',
        'origin': 'Mairie de Perpette-les-oies',
    }
    return eopayment.Payment('dummy', options)


def test_request(backend, freezer):
    freezer.move_to('2020-01-01 00:00:00+01:00')
    transaction_id, method, raw_url = backend.request(
        '10.10', capture_date=datetime.date(2020, 1, 7), subject='Repas pour 4 personnes'
    )
    assert transaction_id
    assert method == 1
    url = urlparse(raw_url)
    assert url.scheme == 'http'
    assert url.netloc == 'dummy-payment.demo.entrouvert.com'
    assert url.path == '/'
    assert url.fragment == ''
    qs = {k: v[0] for k, v in parse_qs(url.query).items()}
    assert qs['transaction_id'] == transaction_id
    assert qs['amount'] == '10.10'
    assert qs['origin'] == 'Mairie de Perpette-les-oies'
    assert qs['capture_day'] == '6'
    assert qs['subject'] == 'Repas pour 4 personnes'


def test_response(backend):
    retour = (
        'http://example.com/retour?amount=10.0'
        '&direct_notification_url=http%3A%2F%2Fexample.com%2Fdirect_notification_url'
        '&email=toto%40example.com'
        '&transaction_id=6Tfw2e1bPyYnz7CedZqvdHt7T9XX6T'
        '&return_url=http%3A%2F%2Fexample.com%2Fretour'
        '&nok=1'
    )
    r = backend.response(retour.split('?', 1)[1])
    assert not r.signed
    assert r.transaction_id == '6Tfw2e1bPyYnz7CedZqvdHt7T9XX6T'
    assert r.return_content is None
    retour = (
        'http://example.com/retour'
        '?amount=10.0'
        '&direct_notification_url=http%3A%2F%2Fexample.com%2Fdirect_notification_url'
        '&email=toto%40example.com'
        '&transaction_id=6Tfw2e1bPyYnz7CedZqvdHt7T9XX6T'
        '&return_url=http%3A%2F%2Fexample.com%2Fretour'
        '&ok=1&signed=1'
    )
    r = backend.response(retour.split('?', 1)[1])
    assert r.signed
    assert r.transaction_id == '6Tfw2e1bPyYnz7CedZqvdHt7T9XX6T'
    assert r.return_content == 'signature ok'

    with pytest.raises(eopayment.ResponseError, match='missing transaction_id'):
        backend.response('foo=bar')



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import importlib
import logging

import pytz

from .common import (  # noqa: F401
    ACCEPTED,
    CANCELED,
    CANCELLED,
    DENIED,
    ERROR,
    EXPIRED,
    FORM,
    HTML,
    PAID,
    RECEIVED,
    URL,
    WAITING,
    BackendNotFound,
    PaymentException,
    ResponseError,
    force_text,
)

__all__ = [
    'Payment',
    'URL',
    'HTML',
    'FORM',
    'SIPS',
    'SYSTEMPAY',
    'TIPI',
    'DUMMY',
    'get_backend',
    'RECEIVED',
    'ACCEPTED',
    'PAID',
    'DENIED',
    'CANCELED',
    'CANCELLED',
    'ERROR',
    'WAITING',
    'EXPIRED',
    'get_backends',
    'PAYFIP_WS',
    'SAGA',
    'KEYWARE',
    'MOLLIE',
]

SIPS = 'sips'
SIPS2 = 'sips2'
SYSTEMPAY = 'systempayv2'
TIPI = 'tipi'
DUMMY = 'dummy'
OGONE = 'ogone'
PAYBOX = 'paybox'
PAYZEN = 'payzen'
PAYFIP_WS = 'payfip_ws'
KEYWARE = 'keyware'
MOLLIE = 'mollie'
SAGA = 'saga'

logger = logging.getLogger(__name__)


def get_backend(kind):
    '''Resolve a backend name into a module object'''
    module = importlib.import_module('.' + kind, package='eopayment')
    return module.Payment


__BACKENDS = [DUMMY, SIPS2, SYSTEMPAY, OGONE, PAYBOX, PAYZEN, TIPI, PAYFIP_WS, KEYWARE, MOLLIE, SAGA]


def get_backends():
    """Return a dictionnary mapping existing eopayment backends name to their
    description.

       >>> get_backends()['dummy'].description['caption']
       'Dummy payment backend'

    """
    return {backend: get_backend(backend) for backend in __BACKENDS}


class Payment:
    """
    Interface to credit card online payment servers of French banks. The
    only use case supported for now is a unique automatic payment.

        >>> options = {
                'numcli': '12345',
            }
        >>> p = Payment(kind=TIPI, options=options)
        >>> transaction_id, kind, data = p.request('10.00', email='bob@example.com')
        >>> print (transaction_id, kind, data) # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        ('...', 1, 'https://www.payfip.gov.fr/tpa/paiement.web?...')

    Supported backend of French banks are:

     - TIPI/PayFiP
     - SIPS 2.0, for BNP, Banque Populaire (before 2010), CCF, HSBC, Crédit
       Agricole, La Banque Postale, LCL, Société Générale and Crédit du
       Nord.
     - SystemPay v2/Payzen for Banque Populaire and Caise d'Epargne (Natixis, after 2010)
     - Ogone
     - Paybox
     - Mollie (Belgium)
     - Keyware (Belgium)

    For SIPs you also need the bank provided middleware especially the two
    executables, request and response, as the protocol from ATOS/SIPS is not
    documented. For the other backends the modules are autonomous.

    Each backend need some configuration parameters to be used, the
    description of the backend list those parameters. The description
    dictionary can be used to generate configuration forms.

        >>> d = get_backend(SPPLUS).description
        >>> print d['caption']
        SPPlus payment service of French bank Caisse d'epargne
        >>> print [p['name'] for p in d['parameters']] # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        ['cle', ..., 'moyen']
        >>> print d['parameters'][0]['caption']
        Secret key, a 40 digits hexadecimal number

    """

    def __init__(self, kind, options, logger=None):
        self.kind = kind
        self.backend = get_backend(kind)(options, logger=logger)

    def request(self, amount, **kwargs):
        """Request a payment to the payment backend.

        Arguments:
        amount -- the amount of money to ask
        email -- the email of the customer (optional)
        usually redundant with the hardwired settings in the bank
        configuration panel. At this url you must use the Payment.response
        method to analyze the bank returned values.

        It returns a triple of values, (transaction_id, kind, data):
          - the first gives a string value to later match the payment with
            the invoice,
          - kind gives the type of the third value, payment.URL or
            payment.HTML or payment.FORM,
          - the third is the URL or the HTML form to contact the payment
            server, which must be sent to the customer browser.
        """
        logger.debug('%r' % kwargs)

        if 'capture_date' in kwargs:
            capture_date = kwargs.pop('capture_date')

            delay_param = False
            for parameter in self.backend.description['parameters']:
                if parameter['name'] == 'capture_day':
                    delay_param = True
                    break

            if not delay_param:
                raise ValueError('capture_date is not supported by the backend.')

            if not isinstance(capture_date, datetime.date):
                raise ValueError('capture_date should be a datetime.date object.')

            # backend timezone should come from some backend configuration
            backend_tz = pytz.timezone('Europe/Paris')
            utc_tz = pytz.timezone('Etc/UTC')
            backend_trans_date = utc_tz.localize(datetime.datetime.utcnow()).astimezone(backend_tz)
            capture_day = (capture_date - backend_trans_date.date()).days
            if capture_day <= 0:
                raise ValueError("capture_date needs to be superior to the transaction date.")

            kwargs['capture_day'] = force_text(capture_day)

        return self.backend.request(amount, **kwargs)

    def response(self, query_string, **kwargs):
        """
        Process a response from the Bank API. It must be used on the URL
        where the user browser of the payment server is going to post the
        result of the payment. Beware it can happen multiple times for the
        same payment, so you MUST support multiple notification of the same
        event, i.e. it should be idempotent. For example if you already
        validated some invoice, receiving a new payment notification for the
        same invoice should alter this state change.

        Beware that when notified directly by the bank (and not through the
        customer browser) no applicative session will exist, so you should
        not depend on it in your handler.

        Arguments:
        query_string -- the URL encoded form-data from a GET or a POST

        It returns a quadruplet of values:

           (result, transaction_id, bank_data, return_content)

         - result is a boolean stating whether the transaction worked, use it
           to decide whether to act on a valid payment,
         - the transaction_id return the same id than returned by request
           when requesting for the payment, use it to find the invoice or
           transaction which is linked to the payment,
         - bank_data is a dictionnary of the data sent by the bank, it should
           be logged for security reasons,
         - return_content, if not None you must return this content as the
           result of the HTTP request, it's used when the bank is calling
           your site as a web service.

        """
        return self.backend.response(query_string, **kwargs)

    def cancel(self, amount, bank_data, **kwargs):
        """
        Cancel or edit the amount of a transaction sent to the bank.

        Arguments:
        - amount -- the amount of money to cancel
        - bank_data -- the transaction dictionary received from the bank
        """
        return self.backend.cancel(amount, bank_data, **kwargs)

    def validate(self, amount, bank_data, **kwargs):
        """
        Validate and trigger the transmission of a transaction to the bank.

        Arguments:
        - amount -- the amount of money
        - bank_data -- the transaction dictionary received from the bank
        """
        return self.backend.validate(amount, bank_data, **kwargs)

    def get_parameters(self, scope='global'):
        res = []
        for param in self.backend.description.get('parameters', []):
            if param.get('scope', 'global') != scope:
                continue
            res.append(param)
        return res

    def get_min_time_between_transactions(self):
        if hasattr(self.backend, 'min_time_between_transactions'):
            return self.backend.min_time_between_transactions
        return 0

    @property
    def has_free_transaction_id(self):
        return self.backend.has_free_transaction_id

    @property
    def has_empty_response(self):
        return self.backend.has_free_transaction_id

    def payment_status(self, transaction_id, **kwargs):
        if not self.backend.payment_status:
            raise NotImplementedError('payment_status is not implemented on this backend')
        return self.backend.payment_status(transaction_id=transaction_id, **kwargs)

    @property
    def has_payment_status(self):
        return hasattr(self.backend, 'payment_status')

    def get_minimal_amount(self):
        return getattr(self.backend, 'minimal_amount', None)

    def get_maximal_amount(self):
        return getattr(self.backend, 'maximal_amount', None)

    @property
    def has_guess(self):
        return hasattr(self.backend.__class__, 'guess')

    @classmethod
    def guess(self, *, method=None, query_string=None, body=None, headers=None, backends=(), **kwargs):
        '''Try to guess the kind of backend and the transaction_id given part of an HTTP response.

        method CAN be GET or POST.
        query_string is the URL encoded query-string as a regular string.
        body is the bytes content of the response.
        headers can eventually give access to the response headers.
        backends is to limit the accepted kind of backends if the possible backends are known.
        '''

        last_exception = None

        for kind, backend in get_backends().items():
            if not hasattr(backend, 'guess'):
                continue
            if backends and kind not in backends:
                continue
            try:
                transaction_id = backend.guess(
                    method=method, query_string=query_string, body=body, headers=headers
                )
            except Exception as e:
                last_exception = e
                continue
            if transaction_id:
                return kind, transaction_id
        if last_exception is not None:
            raise last_exception
        raise BackendNotFound



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from decimal import Decimal

import pytest
import pytz
from six.moves.urllib.parse import parse_qs, urlparse

import eopayment
import eopayment.tipi


def test_tipi():
    p = eopayment.Payment('tipi', {'numcli': '12345'})
    payment_id, kind, url = p.request(
        amount=Decimal('123.12'),
        exer='9999',
        refdet='999900000000999999',
        objet='tout a fait',
        email='info@entrouvert.com',
        urlcl='http://example.com/tipi/test',
        saisie='T',
    )
    assert eopayment.tipi.Payment.REFDET_RE.match(payment_id) is not None
    parsed_qs = parse_qs(urlparse(url).query)
    assert parsed_qs['objet'][0].startswith('tout a fait')
    assert parsed_qs['montant'] == ['12312']
    assert parsed_qs['saisie'] == ['T']
    assert parsed_qs['mel'] == ['info@entrouvert.com']
    assert parsed_qs['numcli'] == ['12345']
    assert parsed_qs['exer'] == ['9999']
    assert parsed_qs['refdet'] == ['999900000000999999']

    response = p.response(
        'objet=tout+a+fait&montant=12312&saisie=T&mel=info%40entrouvert.com'
        '&numcli=12345&exer=9999&refdet=999900000000999999&resultrans=P'
    )
    assert response.signed  # ...
    assert response.order_id == '999900000000999999'
    assert response.transaction_id == '999900000000999999'
    assert response.result == eopayment.PAID

    with pytest.raises(eopayment.ResponseError, match='missing refdet or resultrans'):
        p.response('foo=bar')


def test_tipi_no_orderid_no_refdet():
    p = eopayment.Payment('tipi', {'numcli': '12345'})
    payment_id, kind, url = p.request(
        amount=Decimal('123.12'), exer=9999, email='info@entrouvert.com', saisie='T'
    )
    assert eopayment.tipi.Payment.REFDET_RE.match(payment_id) is not None
    parsed_qs = parse_qs(urlparse(url).query)
    assert 'objet' not in parsed_qs
    assert parsed_qs['montant'] == ['12312']
    assert parsed_qs['saisie'] == ['T']
    assert parsed_qs['mel'] == ['info@entrouvert.com']
    assert parsed_qs['numcli'] == ['12345']
    assert parsed_qs['exer'] == ['9999']
    assert parsed_qs['refdet'][0].startswith(
        datetime.datetime.now(pytz.timezone('Europe/Paris')).strftime('%Y%m%d')
    )


def test_tipi_orderid_refdef_compatible():
    p = eopayment.Payment('tipi', {'numcli': '12345', 'saisie': 'A'})
    payment_id, kind, url = p.request(
        amount=Decimal('123.12'), email='info@entrouvert.com', orderid='F121212'
    )
    assert eopayment.tipi.Payment.REFDET_RE.match(payment_id)
    expected_url = urlparse(eopayment.tipi.TIPI_URL)
    parsed_url = urlparse(url)
    assert parsed_url[:3] == expected_url[:3]
    parsed_qs = parse_qs(parsed_url.query)
    assert 'objet' not in parsed_qs
    assert 'exer' not in parsed_qs
    assert parsed_qs['montant'] == ['12312']
    assert parsed_qs['saisie'] == ['A']
    assert parsed_qs['mel'] == ['info@entrouvert.com']
    assert parsed_qs['numcli'] == ['12345']
    assert parsed_qs['refdet'] == ['F121212']


def test_tipi_orderid_not_refdef_compatible():
    p = eopayment.Payment('tipi', {'numcli': '12345', 'saisie': 'A'})
    payment_id, kind, url = p.request(
        amount=Decimal('123.12'), email='info@entrouvert.com', objet='coucou', orderid='F12-12-12'
    )
    assert eopayment.tipi.Payment.REFDET_RE.match(payment_id) is not None
    expected_url = urlparse(eopayment.tipi.TIPI_URL)
    parsed_url = urlparse(url)
    assert parsed_url[:3] == expected_url[:3]
    parsed_qs = parse_qs(parsed_url.query)
    assert 'exer' not in parsed_qs
    assert parsed_qs['montant'] == ['12312']
    assert parsed_qs['saisie'] == ['A']
    assert parsed_qs['mel'] == ['info@entrouvert.com']
    assert parsed_qs['numcli'] == ['12345']
    assert parsed_qs['refdet'][0].startswith(datetime.datetime.now().strftime('%Y%m%d'))
    assert 'coucou' in parsed_qs['objet'][0]
    assert 'F12-12-12' in parsed_qs['objet'][0]


@pytest.fixture
def payment():
    return eopayment.Payment('tipi', {'numcli': '12345'})


def test_get_minimal_amount(payment):
    assert payment.get_minimal_amount() is not None


def test_get_maximal_amount(payment):
    assert payment.get_maximal_amount() is not None



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import copy
import datetime
import decimal
import functools
import os
import random
import re
import unicodedata
import xml.etree.ElementTree as ET

import pytz
import requests
import zeep
import zeep.exceptions
from six.moves.urllib.parse import parse_qs

from .common import (
    CANCELLED,
    DENIED,
    ERROR,
    EXPIRED,
    PAID,
    URL,
    WAITING,
    PaymentCommon,
    PaymentException,
    PaymentResponse,
    ResponseError,
    _,
    force_text,
)
from .systempayv2 import isonow

# The URL of the WSDL published in the documentation is still wrong, it
# references XSD files which are not resolvable :/ we must use this other URL
# where the XSD files are resolvable. To not depend too much on those files, we
# provide copy in eopayment and we patch them to fix the service binding URL
# and the path of XSD files.  The PayFiP development team is full of morons.
WSDL_URL = 'https://www.payfip.gouv.fr/tpa/services/mas_securite/contrat_paiement_securise/PaiementSecuriseService?wsdl'  # noqa: E501
SERVICE_URL = 'https://www.payfip.gouv.fr/tpa/services/securite'  # noqa: E501
PAYMENT_URL = 'https://www.payfip.gouv.fr/tpa/paiementws.web'

REFDET_RE = re.compile(r'^[A-Za-z0-9]{1,30}$')


def clear_namespace(element):
    def helper(element):
        if element.tag.startswith('{'):
            element.tag = element.tag[element.tag.index('}') + 1 :]
        for subelement in element:
            helper(subelement)

    element = copy.deepcopy(element)
    helper(element)
    return element


def normalize_objet(objet):
    '''Make objet a string of 100 chars in alphabet [A-Za-z0-9 ]'''
    if not objet:
        return objet

    objet = force_text(objet)
    objet = unicodedata.normalize('NFKD', objet).encode('ascii', 'ignore').decode()
    objet = re.sub(r'[\'-]', ' ', objet).strip()
    objet = re.sub(r'[^A-Za-z0-9 ]', '', objet).strip()
    objet = re.sub(r'[\s]+', ' ', objet)
    return objet[:100]


class PayFiPError(PaymentException):
    def __init__(self, code, message, origin=None):
        self.code = code
        self.message = message
        self.origin = origin
        args = [code, message]
        if origin:
            args.append(origin)
        super().__init__(*args)


class PayFiP:
    '''Encapsulate SOAP web-services of PayFiP'''

    def __init__(self, wsdl_url=None, service_url=None, zeep_client_kwargs=None, use_local_wsdl=True):
        # use cached WSDL
        if (not wsdl_url or (wsdl_url == WSDL_URL)) and use_local_wsdl:
            base_path = os.path.join(os.path.dirname(__file__), 'resource', 'PaiementSecuriseService.wsdl')
            wsdl_url = 'file://%s' % base_path
        self.wsdl_url = wsdl_url
        self.service_url = service_url
        self.zeep_client_kwargs = zeep_client_kwargs

    @property
    def client(self):
        if not hasattr(self, '_client'):
            try:
                self._client = zeep.Client(self.wsdl_url or WSDL_URL, **(self.zeep_client_kwargs or {}))
                # distribued WSDL is wrong :/
                self._client.service._binding_options['address'] = self.service_url or SERVICE_URL
            except Exception as e:
                raise PayFiPError('Cound not initialize the SOAP client', e)
        return self._client

    def fault_to_exception(self, fault):
        if (
            fault.message != 'fr.gouv.finances.cp.tpa.webservice.exceptions.FonctionnelleErreur'
            or fault.detail is None
        ):
            return
        detail = clear_namespace(fault.detail)
        code = detail.find('FonctionnelleErreur/code')
        if code is None or not code.text:
            return PayFiPError('inconnu', ET.tostring(detail))
        descriptif = detail.find('FonctionnelleErreur/descriptif')
        libelle = detail.find('FonctionnelleErreur/libelle')
        return PayFiPError(
            code=code.text,
            message=(descriptif is not None and descriptif.text)
            or (libelle is not None and libelle.text)
            or '',
        )

    def _perform(self, request_qname, operation, **kwargs):
        RequestType = self.client.get_type(request_qname)  # noqa: E501
        try:
            return getattr(self.client.service, operation)(RequestType(**kwargs))
        except zeep.exceptions.Fault as fault:
            raise self.fault_to_exception(fault) or PayFiPError('unknown', fault.message, fault)
        except zeep.exceptions.Error as zeep_error:
            raise PayFiPError('SOAP error', str(zeep_error), zeep_error)
        except requests.RequestException as e:
            raise PayFiPError('HTTP error', e)
        except Exception as e:
            raise PayFiPError('Unexpected error', e)

    def get_info_client(self, numcli):
        return self._perform(
            '{http://securite.service.tpa.cp.finances.gouv.fr/reponse}RecupererDetailClientRequest',
            'recupererDetailClient',
            numCli=numcli,
        )

    def get_idop(
        self, numcli, saisie, exer, refdet, montant, mel, url_notification, url_redirect, objet=None
    ):
        return self._perform(
            '{http://securite.service.tpa.cp.finances.gouv.fr/requete}CreerPaiementSecuriseRequest',
            'creerPaiementSecurise',
            numcli=numcli,
            saisie=saisie,
            exer=exer,
            montant=montant,
            refdet=refdet,
            mel=mel,
            urlnotif=url_notification,
            urlredirect=url_redirect,
            objet=objet,
        )

    def get_info_paiement(self, idop):
        return self._perform(
            '{http://securite.service.tpa.cp.finances.gouv.fr/reponse}RecupererDetailPaiementSecuriseRequest',
            'recupererDetailPaiementSecurise',
            idOp=idop,
        )


class Payment(PaymentCommon):
    """Produce requests for and verify response from the TIPI online payment
    processor from the French Finance Ministry.

    """

    description = {
        'caption': 'TIPI, Titres Payables par Internet',
        'parameters': [
            {
                'name': 'numcli',
                'caption': _('Client number'),
                'help_text': _('6 digits number provided by DGFIP'),
                'validation': lambda s: str.isdigit(s) and len(s) == 6,
                'required': True,
            },
            {
                'name': 'saisie',
                'caption': _('Payment type'),
                'default': 'T',
                'choices': [
                    ('T', _('test')),
                    ('X', _('activation')),
                    ('W', _('production')),
                ],
            },
            {
                'name': 'normal_return_url',
                'caption': _('User return URL'),
                'required': True,
            },
            {
                'name': 'automatic_return_url',
                'caption': _('Asynchronous return URL'),
                'required': True,
            },
        ],
    }

    min_time_between_transactions = 60 * 20  # 20 minutes

    minimal_amount = decimal.Decimal('1.0')
    maximal_amount = decimal.Decimal('100000.0')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.payfip = PayFiP()

    def _generate_refdet(self):
        return '%s%010d' % (isonow(), random.randint(1, 1000000000))

    def request(
        self, amount, email, refdet=None, exer=None, orderid=None, subject=None, transaction_id=None, **kwargs
    ):
        montant = self.clean_amount(amount, max_amount=100000)

        numcli = self.numcli
        urlnotif = self.automatic_return_url
        urlredirect = self.normal_return_url

        if not exer:
            exer = str(datetime.date.today().year)

        if refdet:
            pass
        elif transaction_id and REFDET_RE.match(transaction_id):
            refdet = transaction_id
        elif orderid and REFDET_RE.match(orderid):
            refdet = orderid
        else:
            refdet = self._generate_refdet()

        objet_parts = []
        if orderid and refdet != orderid:
            objet_parts.extend(['O', orderid])
        if subject:
            if objet_parts:
                objet_parts.append('S')
            objet_parts.append(subject)
        if transaction_id and refdet != transaction_id:
            objet_parts.extend(['T', transaction_id])
        objet = normalize_objet(' '.join(objet_parts))

        mel = email
        if hasattr(mel, 'decode'):
            mel = email.decode('ascii')
        try:
            if '@' not in mel:
                raise ValueError('no @ in MEL')
            if not (6 <= len(mel) <= 80):
                raise ValueError('len(MEL) is invalid, must be between 6 and 80')
        except Exception as e:
            raise ValueError('MEL is not a valid email, %r' % mel, e)

        # check saisie
        saisie = self.saisie
        if saisie not in ('T', 'X', 'W'):
            raise ValueError('SAISIE invalid format, %r, must be M, T, X or A' % saisie)

        idop = self.payfip.get_idop(
            numcli=numcli,
            saisie=saisie,
            exer=exer,
            refdet=refdet,
            montant=montant,
            mel=mel,
            objet=objet or None,
            url_notification=urlnotif,
            url_redirect=urlredirect,
        )

        return str(idop), URL, PAYMENT_URL + '?idop=%s' % idop

    def payment_status(self, transaction_id, transaction_date=None, **kwargs):
        # idop are valid for 15 minutes after their generation
        # between generation and payment, any call to get_info_paiement() will return a PayFiPError with code=P5
        # before the end of the 15 minutes it can mean the payment is in progress
        # after the 15 minutes period it means the payment will never happen,
        # and after one day the code will change for P1, meaning the idop is
        # now unknown as it as been cleaned by the night cleaning job.
        #
        # So in order to interpret the meaning of PayFiP error codes we need
        # the date of the start of the transaction and add to it some margin
        # to.
        idop = transaction_id
        if transaction_date:
            if transaction_date.tzinfo:  # date is aware
                now = datetime.datetime.now(tz=pytz.utc)
            else:
                now = datetime.datetime.now()
            delta = now - transaction_date
        else:
            delta = datetime.timedelta(seconds=0)
        # set the threshold between transaction 'in progress' and 'expired' at 20 minutes
        threshold = datetime.timedelta(seconds=20 * 60)

        try:
            response = self.payfip.get_info_paiement(idop)
        except PayFiPError as e:
            if e.code == 'P1' or (e.code == 'P5' and delta >= threshold):
                return PaymentResponse(result=EXPIRED, signed=True, order_id=transaction_id)
            if e.code == 'P5' and delta < threshold:
                return PaymentResponse(result=WAITING, signed=True, order_id=transaction_id)
            raise e
        return self.payfip_response_to_eopayment_response(idop, response)

    def response(self, query_string, **kwargs):
        fields = parse_qs(query_string, True)
        idop = (fields.get('idop') or [None])[0]

        if not idop:
            raise ResponseError('missing idop parameter in query string')

        return self.payment_status(idop)

    @classmethod
    def payfip_response_to_eopayment_response(cls, idop, response):
        if response.resultrans == 'P':
            result = PAID
            bank_status = 'paid CB'
        elif response.resultrans == 'V':
            result = PAID
            bank_status = 'paid direct debit'
        elif response.resultrans == 'R':
            result = DENIED
            bank_status = 'refused CB'
        elif response.resultrans == 'Z':
            result = DENIED
            bank_status = 'refused direct debit'
        elif response.resultrans == 'A':
            result = CANCELLED
            bank_status = 'cancelled CB'
        else:
            result = ERROR
            bank_status = 'unknown result code: %r' % response.resultrans

        transaction_id = response.refdet
        transaction_id += ' ' + idop
        if response.numauto:
            transaction_id += ' ' + response.numauto

        return PaymentResponse(
            result=result,
            bank_status=bank_status,
            signed=True,
            bank_data={k: response[k] for k in response},
            order_id=idop,
            transaction_id=transaction_id,
            test=response.saisie == 'T',
        )

    @classmethod
    def guess(self, *, method=None, query_string=None, body=None, headers=None, backends=(), **kwargs):
        for content in [query_string, body]:
            if isinstance(content, bytes):
                try:
                    content = content.decode()
                except UnicodeDecodeError:
                    pass
            if isinstance(content, str):
                fields = parse_qs(content)
                if set(fields) == {'idOp'}:
                    return fields['idOp'][0]
        return None


if __name__ == '__main__':
    import click

    def show_payfip_error(func):
        @functools.wraps(func)
        def f(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except PayFiPError as e:
                click.echo(click.style('PayFiP ERROR : %s "%s"' % (e.code, e.message), fg='red'))

        return f

    @click.group()
    @click.option('--wsdl-url', default=None)
    @click.option('--service-url', default=None)
    @click.pass_context
    def main(ctx, wsdl_url, service_url):
        import logging

        logging.basicConfig(level=logging.INFO)
        # hide warning from zeep
        logging.getLogger('zeep.wsdl.bindings.soap').level = logging.ERROR

        ctx.obj = PayFiP(wsdl_url=wsdl_url, service_url=service_url)

    def numcli(ctx, param, value):
        if not isinstance(value, str) or len(value) != 6 or not value.isdigit():
            raise click.BadParameter('numcli must a 6 digits number')
        return value

    @main.command()
    @click.argument('numcli', callback=numcli, type=str)
    @click.pass_obj
    @show_payfip_error
    def info_client(payfip, numcli):
        response = payfip.get_info_client(numcli)
        for key in response:
            print('%15s:' % key, response[key])

    @main.command()
    @click.argument('numcli', callback=numcli, type=str)
    @click.option('--saisie', type=click.Choice(['T', 'X', 'W']), required=True)
    @click.option('--exer', type=str, required=True)
    @click.option('--montant', type=int, required=True)
    @click.option('--refdet', type=str, required=True)
    @click.option('--mel', type=str, required=True)
    @click.option('--url-notification', type=str, required=True)
    @click.option('--url-redirect', type=str, required=True)
    @click.option('--objet', default=None, type=str)
    @click.pass_obj
    @show_payfip_error
    def get_idop(payfip, numcli, saisie, exer, montant, refdet, mel, objet, url_notification, url_redirect):
        idop = payfip.get_idop(
            numcli=numcli,
            saisie=saisie,
            exer=exer,
            montant=montant,
            refdet=refdet,
            mel=mel,
            objet=objet,
            url_notification=url_notification,
            url_redirect=url_redirect,
        )
        print('idOp:', idop)
        print(PAYMENT_URL + '?idop=%s' % idop)

    @main.command()
    @click.argument('idop', type=str)
    @click.pass_obj
    @show_payfip_error
    def info_paiement(payfip, idop):
        print(payfip.get_info_paiement(idop))

    main()



# eopayment - online payment library
# Copyright (C) 2011-2022 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

import eopayment


def test_get_backends():
    assert len(eopayment.get_backends()) > 1


GUESS_TEST_VECTORS = [
    {
        'name': 'tipi',
        'kwargs': {
            'query_string': 'objet=tout+a+fait&montant=12312&saisie=T&mel=info%40entrouvert.com'
            '&numcli=12345&exer=9999&refdet=999900000000999999&resultrans=P',
        },
        'result': ['tipi', '999900000000999999'],
    },
    {
        'name': 'payfip_ws',
        'kwargs': {
            'query_string': 'idOp=1234',
        },
        'result': ['payfip_ws', '1234'],
    },
    {
        'name': 'systempayv2-old-transaction-id',
        'kwargs': {
            'query_string': 'vads_amount=1042&vads_auth_mode=FULL&vads_auth_number=3feadf'
            '&vads_auth_result=00&vads_capture_delay=0&vads_card_brand=CB'
            '&vads_result=00'
            '&vads_card_number=497010XXXXXX0000'
            '&vads_payment_certificate=582ba2b725057618706d7a06e9e59acdbf69ff53'
            '&vads_ctx_mode=TEST&vads_currency=978&vads_effective_amount=1042'
            '&vads_site_id=70168983&vads_trans_date=20161013101355'
            '&vads_trans_id=226787&vads_trans_uuid=4b5053b3b1fe4b02a07753e7a'
            '&vads_effective_creation_date=20200330162530'
            '&signature=c17fab393f94dc027dc029510c85d5fc46c4710f',
        },
        'result': ['systempayv2', '20161013101355_226787'],
    },
    {
        'name': 'systempayv2-eo-trans-id',
        'kwargs': {
            'query_string': 'vads_amount=1042&vads_auth_mode=FULL&vads_auth_number=3feadf'
            '&vads_auth_result=00&vads_capture_delay=0&vads_card_brand=CB'
            '&vads_result=00'
            '&vads_card_number=497010XXXXXX0000'
            '&vads_payment_certificate=582ba2b725057618706d7a06e9e59acdbf69ff53'
            '&vads_ctx_mode=TEST&vads_currency=978&vads_effective_amount=1042'
            '&vads_site_id=70168983&vads_trans_date=20161013101355'
            '&vads_trans_id=226787&vads_trans_uuid=4b5053b3b1fe4b02a07753e7a'
            '&vads_effective_creation_date=20200330162530'
            '&signature=c17fab393f94dc027dc029510c85d5fc46c4710f'
            '&vads_ext_info_eopayment_trans_id=123456',
        },
        'result': ['systempayv2', '123456'],
    },
    {
        'name': 'paybox',
        'kwargs': {
            'query_string': 'montant=4242&reference=abcdef&code_autorisation=A'
            '&erreur=00000&date_transaction=20200101&heure_transaction=01%3A01%3A01',
        },
        'result': ['paybox', 'abcdef'],
    },
    {
        'name': 'ogone-no-complus',
        'kwargs': {
            'query_string': 'orderid=myorder&status=9&payid=3011229363&cn=Us%C3%A9r&ncerror=0'
            '&trxdate=10%2F24%2F16&acceptance=test123&currency=eur&amount=7.5',
        },
        'result': ['ogone', 'myorder'],
    },
    {
        'name': 'ogone-with-complus',
        'kwargs': {
            'query_string': 'complus=neworder&orderid=myorder&status=9&payid=3011229363&cn=Us%C3%A9r'
            '&ncerror=0&trxdate=10%2F24%2F16&acceptance=test123&currency=eur&amount=7.5',
        },
        'result': ['ogone', 'neworder'],
    },
    {
        'name': 'mollie',
        'kwargs': {
            'body': b'id=tr_7UhSN1zuXS',
        },
        'result': ['mollie', 'tr_7UhSN1zuXS'],
    },
    {
        'name': 'sips2',
        'kwargs': {
            'body': (
                b'Data=captureDay%3D0%7CcaptureMode%3DAUTHOR_CAPTURE%7CcurrencyCode%3D978%7CmerchantId%3D002001000000001%7CorderChannel%3D'
                b'INTERNET%7CresponseCode%3D00%7CtransactionDateTime%3D2016-02-01T17%3A44%3A20%2B01%3A00%7C'
                b'transactionReference%3D668930%7CkeyVersion%3D1%7CacquirerResponseCode%3D00%7Camou'
                b'nt%3D1200%7CauthorisationId%3D12345%7CcardCSCResultCode%3D4E%7CpanExpiryDate%3D201605%7Cpay'
                b'mentMeanBrand%3DMASTERCARD%7CpaymentMeanType%3DCARD%7CcustomerIpAddress%3D82.244.203.243%7CmaskedPan'
                b'%3D5100%23%23%23%23%23%23%23%23%23%23%23%2300%7CorderId%3Dd4903de7027f4d56ac01634fd7ab9526%7CholderAuthentRelegation'
                b'%3DN%7CholderAuthentStatus%3D3D_ERROR%7CtransactionOrigin%3DINTERNET%7CpaymentPattern%3D'
                b'ONE_SHOT&Seal=6ca3247765a19b45d25ad54ef4076483e7d55583166bd5ac9c64357aac097602&InterfaceVersion=HP_2.0&Encode='
            ),
        },
        'result': ['sips2', '668930'],
    },
    {
        'name': 'notfound',
        'kwargs': {},
        'exception': eopayment.BackendNotFound,
    },
    {
        'name': 'notfound-2',
        'kwargs': {'query_string': None, 'body': [12323], 'headers': {b'1': '2'}},
        'exception': eopayment.BackendNotFound,
    },
    {
        'name': 'backends-limitation',
        'kwargs': {
            'body': b'id=tr_7UhSN1zuXS',
            'backends': ['payfips_ws'],
        },
        'exception': eopayment.BackendNotFound,
    },
]


@pytest.mark.parametrize('test_vector', GUESS_TEST_VECTORS, ids=lambda tv: tv['name'])
def test_guess(test_vector):
    kwargs, result, exception = test_vector['kwargs'], test_vector.get('result'), test_vector.get('exception')
    if exception is not None:
        with pytest.raises(exception):
            eopayment.Payment.guess(**kwargs)
    else:
        assert list(eopayment.Payment.guess(**kwargs)) == result



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

import pytest
import requests
from httmock import HTTMock, all_requests, remember_called, response, urlmatch, with_httmock

import eopayment
from eopayment.keyware import Payment

WEBHOOK_URL = 'https://callback.example.com'
RETURN_URL = 'https://return.example.com'
API_KEY = 'test'

ORDER_ID = "1c969951-f5f1-4290-ae41-6177961fb3cb"
QUERY_STRING = 'order_id=' + ORDER_ID


POST_ORDER_RESPONSE = {
    "amount": 995,
    "client": {"user_agent": "Testing API"},
    "created": "2016-07-04T11:41:57.121017+00:00",
    "currency": "EUR",
    "description": "Example description",
    "id": ORDER_ID,
    "merchant_id": "7131b462-1b7d-489f-aba9-de2f0eadc9dc",
    "modified": "2016-07-04T11:41:57.183822+00:00",
    "order_url": "https://api.online.emspay.eu/pay/1c969951-f5f1-4290-ae41-6177961fb3cb/",
    "project_id": "1ef558ed-d77d-470d-b43b-c0f4a131bcef",
    "status": "new",
}

GET_ORDER_RESPONSE = {
    "amount": 995,
    "client": {"user_agent": "Testing API"},
    "created": "2016-07-04T11:41:55.635115+00:00",
    "currency": "EUR",
    "description": "Example order #1",
    "id": ORDER_ID,
    "last_transaction_added": "2016-07-04T11:41:55.831655+00:00",
    "merchant_id": "7131b462-1b7d-489f-aba9-de2f0eadc9dc",
    "merchant_order_id": "EXAMPLE001",
    "modified": "2016-07-04T11:41:56.215543+00:00",
    "project_id": "1ef558ed-d77d-470d-b43b-c0f4a131bcef",
    "return_url": "http://www.example.com/",
    "status": "completed",
    "transactions": [
        {
            "amount": 995,
            "balance": "internal",
            "created": "2016-07-04T11:41:55.831655+00:00",
            "credit_debit": "credit",
            "currency": "EUR",
            "description": "Example order #1",
            "events": [
                {
                    "event": "new",
                    "id": "0c4bd0cd-f197-446b-b218-39cbeb028290",
                    "noticed": "2016-07-04T11:41:55.987468+00:00",
                    "occurred": "2016-07-04T11:41:55.831655+00:00",
                    "source": "set_status",
                }
            ],
            "expiration_period": "PT60M",
            "id": "6c81499c-14e4-4974-99e5-fe72ce019411",
            "merchant_id": "7131b462-1b7d-489f-aba9-de2f0eadc9dc",
            "modified": "2016-07-04T11:41:56.065147+00:00",
            "order_id": ORDER_ID,
            "payment_method": "ideal",
            "payment_method_details": {
                "issuer_id": "INGBNL2A",
            },
            "payment_url": "https://api.online.emspay.eu/redirect/6c81499c-14e4-4974-99e5-fe72ce019411/to/payment/",
            "project_id": "1ef558ed-d77d-470d-b43b-c0f4a131bcef",
            "status": "completed",
        }
    ],
}


@remember_called
@urlmatch(scheme='https', netloc='api.online.emspay.eu', path='/v1/orders', method='POST')
def add_order(url, request):
    return response(200, POST_ORDER_RESPONSE, request=request)


@remember_called
@urlmatch(scheme='https', netloc='api.online.emspay.eu', path='/v1/orders', method='GET')
def successful_order(url, request):
    return response(200, GET_ORDER_RESPONSE, request=request)


@remember_called
@urlmatch(scheme='https', netloc='api.online.emspay.eu', path=r'/v1/orders', method='DELETE')
def cancelled_order(url, request):
    resp = GET_ORDER_RESPONSE.copy()
    resp['status'] = 'cancelled'
    return response(200, resp, request=request)


@remember_called
@urlmatch(scheme='https', netloc='api.online.emspay.eu', path='/v1/orders')
def error_order(url, request):
    resp = GET_ORDER_RESPONSE.copy()
    resp['status'] = 'error'
    return response(200, resp, request=request)


@remember_called
@urlmatch(scheme='https', netloc='api.online.emspay.eu', path='/v1/orders', method='GET')
def connection_error(url, request):
    raise requests.ConnectionError('test msg')


@remember_called
@urlmatch(scheme='https', netloc='api.online.emspay.eu', path='/v1/orders', method='GET')
def http_error(url, request):
    error_payload = {'error': {'status': 400, 'type': 'Bad request', 'value': 'error'}}
    return response(400, error_payload, request=request)


@remember_called
@urlmatch(scheme='https', netloc='api.online.emspay.eu', path='/v1/orders', method='GET')
def invalid_json(url, request):
    return response(200, '{', request=request)


@pytest.fixture
def keyware():
    return Payment(
        {
            'normal_return_url': RETURN_URL,
            'automatic_return_url': WEBHOOK_URL,
            'api_key': API_KEY,
        }
    )


@with_httmock(add_order)
def test_keyware_request(keyware):
    email = 'test@test.com'
    order_id, kind, url = keyware.request(2.5, email=email)

    assert order_id == ORDER_ID
    assert kind == eopayment.URL
    assert 'api.online.emspay.eu/pay/' in url

    body = json.loads(add_order.call['requests'][0].body.decode())
    assert body['currency'] == 'EUR'
    assert body['customer']['email_address'] == email
    assert isinstance(body['amount'], int)
    assert body['amount'] == 250
    assert body['webhook_url'] == WEBHOOK_URL
    assert body['return_url'] == RETURN_URL


@with_httmock(successful_order)
def test_keyware_response(keyware):
    payment_response = keyware.response(QUERY_STRING)

    assert payment_response.result == eopayment.PAID
    assert payment_response.signed is True
    assert payment_response.bank_data == GET_ORDER_RESPONSE
    assert payment_response.order_id == ORDER_ID
    assert payment_response.transaction_id == ORDER_ID
    assert payment_response.bank_status == 'completed'
    assert payment_response.test is False

    request = successful_order.call['requests'][0]
    assert ORDER_ID in request.url


@with_httmock(error_order)
def test_keyware_response_error(keyware):
    payment_response = keyware.response(QUERY_STRING)
    assert payment_response.result == eopayment.ERROR


@with_httmock(cancelled_order)
def test_keyware_cancel(keyware):
    resp = keyware.cancel(amount=995, bank_data=POST_ORDER_RESPONSE)
    request = cancelled_order.call['requests'][0]
    assert ORDER_ID in request.url


@with_httmock(error_order)
def test_keyware_cancel_error(keyware):
    with pytest.raises(eopayment.ResponseError) as excinfo:
        keyware.cancel(amount=995, bank_data=POST_ORDER_RESPONSE)
        assert 'expected "cancelled" status, got "error" instead' in str(excinfo.value)


@with_httmock(connection_error)
def test_keyware_endpoint_connection_error(keyware):
    with pytest.raises(eopayment.PaymentException) as excinfo:
        keyware.call_endpoint('GET', 'orders')
        assert 'test msg' in str(excinfo.value)


@with_httmock(http_error)
def test_keyware_endpoint_http_error(keyware):
    with pytest.raises(eopayment.PaymentException) as excinfo:
        keyware.call_endpoint('GET', 'orders')
        assert 'Bad request' in str(excinfo.value)


@with_httmock(invalid_json)
def test_keyware_endpoint_json_error(keyware):
    with pytest.raises(eopayment.PaymentException) as excinfo:
        keyware.call_endpoint('GET', 'orders')
        assert 'JSON' in str(excinfo.value)



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import configparser
import decimal
import logging
import os.path
import pprint
import subprocess
import tempfile

import click

from . import FORM, URL, Payment


def option(value):
    try:
        name, value = value.split('=', 1)
    except Exception:
        raise ValueError('invalid option %s' % value)
    return (name, value)


@click.group()
@click.option('--option', type=option, multiple=True)
@click.option('--name', type=str, default='')
@click.option('--backend', type=str)
@click.option('--debug/--no-debug', default=None)
@click.pass_context
def main(ctx, backend, debug, option, name):
    config_file = os.path.expanduser('~/.config/eopayment.ini')
    option = list(option)
    if os.path.exists(config_file):
        parser = configparser.ConfigParser(interpolation=None)
        with open(config_file) as fd:
            parser.read_file(fd)
        if debug is None:
            debug = parser.getboolean('default', 'debug', fallback=False)
        if debug is True:
            logging.basicConfig(level=logging.DEBUG)
        for section in parser.sections():
            if section == 'default':
                continue
            if ':' in section:
                config_backend, config_name = section.split(':', 1)
            else:
                config_backend, config_name = section, None
            load = False
            if not name and not backend:
                # use first section
                logging.debug('no backend and not name given using first section found')
                backend = config_backend
                load = True
            elif name and backend:
                load = config_backend == backend and config_name == name
            elif name:
                load = config_name == name
            elif backend:
                load = config_backend == backend
            if load:
                logging.debug('loading configuration "%s"', section)
                backend = backend or config_backend
                option.extend(parser.items(section=section))
                break
        else:
            if not backend:
                raise ValueError('no backend found')

        if parser.has_section(backend):
            option.extend(parser.items(section=backend))
    if debug is True:
        logging.basicConfig(level=logging.DEBUG)
    ctx.obj = Payment(backend, dict(option))


@main.command()
@click.argument('amount', type=decimal.Decimal)
@click.option('--param', type=option, multiple=True)
@click.pass_obj
def request(backend, amount, param):
    transaction_id, kind, what = backend.request(amount, **dict(param))
    print('Transaction ID:', transaction_id)
    if kind == FORM:
        print(what)
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            print('<!doctype html> <html> <head> <meta charset="utf-8"> </head> <body>', file=fd)
            print(what, file=fd)
            print('''<script>document.querySelector('input[type="submit"]').click()</script>''', file=fd)
            subprocess.call(['gnome-www-browser', fd.name])
            print('</body> <html>', file=fd)
    elif kind == URL:
        print('Please click on URL:', what)
        subprocess.call(['gnome-www-browser', what])


@main.command()
@click.argument('query_string', type=str)
@click.option('--param', type=option, multiple=True)
@click.pass_obj
def response(backend, query_string, param):
    payment_response = backend.response(query_string, **dict(param))
    for key, value in payment_response.__dict__.items():
        if not isinstance(value, (dict, list)):
            print(' ', key, ':', value)
        else:
            print(' ', key, ':')
            formatted_value = pprint.pformat(value)
            for line in formatted_value.splitlines(False):
                print(' ', line)


main()



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import base64
import codecs
from decimal import Decimal
from unittest import TestCase, mock
from xml.etree import ElementTree as ET

import pytest
from six.moves.urllib import parse as urllib

import eopayment
import eopayment.paybox as paybox

BACKEND_PARAMS = {
    'platform': 'test',
    'site': '12345678',
    'rang': '001',
    'identifiant': '12345678',
    'shared_secret': (
        '0123456789ABCDEF0123456789ABCDEF01234'
        '56789ABCDEF0123456789ABCDEF0123456789'
        'ABCDEF0123456789ABCDEF0123456789ABCDE'
        'F0123456789ABCDEF'
    ),
    'automatic_return_url': 'http://example.com/callback',
}


class PayboxTests(TestCase):
    def test_sign(self):
        key = (
            b'0123456789ABCDEF0123456789ABCDEF0123456789'
            b'ABCDEF0123456789ABCDEF0123456789ABCDEF0123'
            b'456789ABCDEF0123456789ABCDEF0123456789ABCD'
            b'EF'
        )
        key = codecs.decode(key, 'hex')
        d = dict(
            paybox.sign(
                [
                    ['PBX_SITE', '12345678'],
                    ['PBX_RANG', '32'],
                    ['PBX_IDENTIFIANT', '12345678'],
                    ['PBX_TOTAL', '999'],
                    ['PBX_DEVISE', '978'],
                    ['PBX_CMD', 'appel à Paybox'],
                    ['PBX_PORTEUR', 'test@paybox.com'],
                    ['PBX_RETOUR', 'Mt:M;Ref:R;Auto:A;Erreur:E'],
                    ['PBX_HASH', 'SHA512'],
                    ['PBX_TIME', '2015-06-08T16:21:16+02:00'],
                ],
                key,
            )
        )
        result = (
            '7E74D8E9A0DBB65AAE51C5C50C2668FD98FC99AED'
            'F18244BB1935F602B6C2E953B61FD84364F34FDB8'
            '8B049901C0A07F6040AF446BBF5589113F48A733D'
            '551D4'
        )
        self.assertIn('PBX_HMAC', d)
        self.assertEqual(d['PBX_HMAC'], result)

    def test_request(self):
        backend = eopayment.Payment('paybox', BACKEND_PARAMS)
        time = '2015-07-15T18:26:32+02:00'
        email = 'bdauvergne@entrouvert.com'
        order_id = '20160216'
        transaction = '1234'
        amount = '19.99'
        transaction_id, kind, what = backend.request(
            Decimal(amount),
            email=email,
            orderid=order_id,
            transaction_id=transaction,
            time=time,
            manual_validation=False,
        )
        self.assertEqual(kind, eopayment.FORM)
        self.assertEqual(transaction_id, '%s!%s' % (transaction, order_id))
        root = ET.fromstring(str(what))
        self.assertEqual(root.tag, 'form')
        self.assertEqual(root.attrib['method'], 'POST')
        self.assertEqual(root.attrib['action'], paybox.URLS['test'])
        expected_form_values = {
            'PBX_RANG': '001',
            'PBX_SITE': '12345678',
            'PBX_IDENTIFIANT': '12345678',
            'PBX_RETOUR': (
                'montant:M;reference:R;code_autorisation:A;'
                'erreur:E;numero_appel:T;numero_transaction:S;'
                'date_transaction:W;heure_transaction:Q;signature:K'
            ),
            'PBX_TIME': time,
            'PBX_PORTEUR': email,
            'PBX_CMD': '%s!%s' % (transaction, order_id),
            'PBX_TOTAL': amount.replace('.', ''),
            'PBX_DEVISE': '978',
            'PBX_HASH': 'SHA512',
            'PBX_HMAC': (
                'A9F561A6EA79390F1741A6B72872470BC1A1688E4581'
                'F097EC80B99D2038413AB350F2F5429FFA4F8D426D99'
                'B72E038164642F6F9BA10D46837EE486EEB944A2'
            ),
            'PBX_ARCHIVAGE': '20160216',
            'PBX_REPONDRE_A': 'http://example.com/callback',
            'PBX_AUTOSEULE': 'N',
        }

        form_params = {}
        for node in root:
            self.assertIn(node.attrib['type'], ('hidden', 'submit'))
            if node.attrib['type'] == 'submit':
                self.assertEqual(set(node.attrib.keys()), {'type', 'value'})
            if node.attrib['type'] == 'hidden':
                self.assertEqual(set(node.attrib.keys()), {'type', 'name', 'value'})
                name = node.attrib['name']
                form_params[name] = node.attrib['value']
        assert form_params == expected_form_values

    def test_request_with_capture_day(self):
        params = BACKEND_PARAMS.copy()
        time = '2018-08-21T10:26:32+02:00'
        email = 'user@entrouvert.com'
        order_id = '20180821'
        transaction = '1234'
        amount = '42.99'

        for capture_day in ('7', '07'):
            params['capture_day'] = capture_day
            backend = eopayment.Payment('paybox', params)
            transaction_id, kind, what = backend.request(
                Decimal(amount), email=email, orderid=order_id, transaction_id=transaction, time=time
            )
            root = ET.fromstring(str(what))

            form_params = {
                node.attrib['name']: node.attrib['value'] for node in root if node.attrib['type'] == 'hidden'
            }
            self.assertIn('PBX_DIFF', form_params)
            self.assertEqual(form_params['PBX_DIFF'], '07')

        # capture_day can be used as a request argument
        params = BACKEND_PARAMS.copy()
        backend = eopayment.Payment('paybox', params)
        transaction_id, kind, what = backend.request(
            Decimal(amount),
            email=email,
            orderid=order_id,
            transaction_id=transaction,
            time=time,
            capture_day='2',
        )
        root = ET.fromstring(str(what))

        form_params = {
            node.attrib['name']: node.attrib['value'] for node in root if node.attrib['type'] == 'hidden'
        }
        self.assertIn('PBX_DIFF', form_params)
        self.assertEqual(form_params['PBX_DIFF'], '02')

        # capture_day passed as a request argument
        # overrides capture_day from backend params
        params = BACKEND_PARAMS.copy()
        params['capture_day'] = '7'
        backend = eopayment.Payment('paybox', params)
        transaction_id, kind, what = backend.request(
            Decimal(amount),
            email=email,
            orderid=order_id,
            transaction_id=transaction,
            time=time,
            capture_day='2',
        )
        root = ET.fromstring(str(what))

        form_params = {
            node.attrib['name']: node.attrib['value'] for node in root if node.attrib['type'] == 'hidden'
        }
        self.assertIn('PBX_DIFF', form_params)
        self.assertEqual(form_params['PBX_DIFF'], '02')

    def test_request_with_authorization_only(self):
        params = BACKEND_PARAMS.copy()
        time = '2018-08-21T10:26:32+02:00'
        email = 'user@entrouvert.com'
        order_id = '20180821'
        transaction = '1234'
        amount = '42.99'

        params['capture_mode'] = 'AUTHOR_CAPTURE'
        backend = eopayment.Payment('paybox', params)
        transaction_id, kind, what = backend.request(
            Decimal(amount), email=email, orderid=order_id, transaction_id=transaction, time=time
        )
        root = ET.fromstring(str(what))

        form_params = {
            node.attrib['name']: node.attrib['value'] for node in root if node.attrib['type'] == 'hidden'
        }
        self.assertEqual(form_params['PBX_AUTOSEULE'], 'O')

    def test_response(self):
        backend = eopayment.Payment('paybox', BACKEND_PARAMS)
        order_id = '20160216'
        transaction = '1234'
        reference = '%s!%s' % (transaction, order_id)
        data = {
            'montant': '4242',
            'reference': reference,
            'code_autorisation': 'A',
            'erreur': '00000',
            'date_transaction': '20200101',
            'heure_transaction': '01:01:01',
        }
        response = backend.response(urllib.urlencode(data))
        self.assertEqual(response.order_id, reference)
        assert not response.signed
        assert response.transaction_date.isoformat() == '2020-01-01T01:01:01+01:00'

        with self.assertRaisesRegex(eopayment.ResponseError, 'missing erreur or reference'):
            backend.response('foo=bar')

    def test_perform_operations(self):
        operations = {'validate': '00002', 'cancel': '00055'}
        for operation_name, operation_code in operations.items():
            params = BACKEND_PARAMS.copy()
            params['cle'] = 'cancelling_key'
            backend = eopayment.Payment('paybox', params)
            bank_data = {
                'numero_transaction': ['13957441'],
                'numero_appel': ['30310733'],
                'reference': ['830657461681'],
            }
            backend_raw_response = (
                'NUMTRANS=0013989865&NUMAPPEL=0030378572&NUMQUESTION=0013989862'
                '&SITE=1999888&RANG=32&AUTORISATION=XXXXXX&CODEREPONSE=00000'
                '&COMMENTAIRE=Demande traitée avec succès&REFABONNE=&PORTEUR='
            )
            backend_expected_response = {
                "CODEREPONSE": "00000",
                "RANG": "32",
                "AUTORISATION": "XXXXXX",
                "NUMTRANS": "0013989865",
                "PORTEUR": "",
                "COMMENTAIRE": "Demande traitée avec succès",
                "SITE": "1999888",
                "NUMAPPEL": "0030378572",
                "REFABONNE": "",
                "NUMQUESTION": "0013989862",
            }

            with mock.patch('eopayment.paybox.requests.post') as requests_post:
                response = mock.Mock(status_code=200, text=backend_raw_response)
                requests_post.return_value = response
                backend_response = getattr(backend, operation_name)(Decimal('10'), bank_data)
                self.assertEqual(requests_post.call_args[0][0], 'https://preprod-ppps.paybox.com/PPPS.php')
                params_sent = requests_post.call_args[0][1]
                # make sure the date parameter is present
                assert 'DATEQ' in params_sent
                # don't care about its value
                params_sent.pop('DATEQ')
                expected_params = {
                    'CLE': 'cancelling_key',
                    'VERSION': '00103',
                    'TYPE': operation_code,
                    'MONTANT': '1000',
                    'NUMAPPEL': '30310733',
                    'NUMTRANS': '13957441',
                    'NUMQUESTION': '0013957441',
                    'REFERENCE': '830657461681',
                    'RANG': backend.backend.rang,
                    'SITE': backend.backend.site,
                    'DEVISE': backend.backend.devise,
                }
                self.assertEqual(params_sent, expected_params)
                self.assertEqual(backend_response, backend_expected_response)

                params['platform'] = 'prod'
                backend = eopayment.Payment('paybox', params)
                with mock.patch('eopayment.paybox.requests.post') as requests_post:
                    response = mock.Mock(status_code=200, text=backend_raw_response)
                    requests_post.return_value = response
                    getattr(backend, operation_name)(Decimal('10'), bank_data)
                    self.assertEqual(requests_post.call_args[0][0], 'https://ppps.paybox.com/PPPS.php')

                with mock.patch('eopayment.paybox.requests.post') as requests_post:
                    error_response = """CODEREPONSE=00015&COMMENTAIRE=PAYBOX : Transaction non trouvée"""
                    response = mock.Mock(status_code=200, text=error_response)
                    requests_post.return_value = response
                    self.assertRaisesRegex(
                        eopayment.ResponseError,
                        'Transaction non trouvée',
                        getattr(backend, operation_name),
                        Decimal('10'),
                        bank_data,
                    )

    def test_validate_payment(self):
        params = BACKEND_PARAMS.copy()
        params['cle'] = 'cancelling_key'
        backend = eopayment.Payment('paybox', params)
        bank_data = {
            'numero_transaction': ['13957441'],
            'numero_appel': ['30310733'],
            'reference': ['830657461681'],
        }
        backend_raw_response = (
            'NUMTRANS=0013989865&NUMAPPEL=0030378572&NUMQUESTION=0013989862'
            '&SITE=1999888&RANG=32&AUTORISATION=XXXXXX&CODEREPONSE=00000'
            '&COMMENTAIRE=Demande traitée avec succès&REFABONNE=&PORTEUR='
        )

        with mock.patch('eopayment.paybox.requests.post') as requests_post:
            response = mock.Mock(status_code=200, text=backend_raw_response)
            requests_post.return_value = response
            backend.validate(Decimal(12), bank_data)

    def test_rsa_signature_validation(self):
        pkey = '''-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDUgYufHuheMztK1LhQSG6xsOzb
UX4D2A/QcMvkEcRVXFx5tQqcE9/JnMqE41TF/ebn7jC/MBxxtPFkUN7+EZoeMN7x
OWzAMDm/xsCWRvvel4GGixgm3aQRUPyTrlm4Ksy32Ya0rNnEDMAvB3dxOn7cp8GR
ZdzrudBlevZXpr6iYwIDAQAB
-----END PUBLIC KEY-----'''
        data = 'coin\n'
        sig64 = '''VCt3sgT0ecacmDEWWNVXJ+jGmIPBMApK42tBJV0FlDjpllOGPy8MsAmLW4/QjTtx
z0Dkz0NjxvU+5WzQZh9Uuxr/egRCwV4NMRWqu0zaVVioeBvl4/5CWm4f4/1L9+0m
FBFKOZhgBJnkC+l6+XhT4aYWKaQ4ocmOMV92yjeXTE4='''
        self.assertTrue(paybox.verify(data, base64.b64decode(sig64), key=pkey))

    def test_request_manual_validation(self):
        params = BACKEND_PARAMS.copy()
        time = '2018-08-21T10:26:32+02:00'
        email = 'user@entrouvert.com'
        order_id = '20180821'
        transaction = '1234'
        amount = '42.99'

        backend = eopayment.Payment('paybox', params)

        transaction_id, kind, what = backend.request(
            Decimal(amount), email=email, orderid=order_id, transaction_id=transaction, time=time
        )
        root = ET.fromstring(str(what))
        form_params = {
            node.attrib['name']: node.attrib['value'] for node in root if node.attrib['type'] == 'hidden'
        }
        self.assertIn('PBX_AUTOSEULE', form_params)
        self.assertEqual(form_params['PBX_AUTOSEULE'], 'N')

        transaction_id, kind, what = backend.request(
            Decimal(amount),
            email=email,
            orderid=order_id,
            transaction_id=transaction,
            time=time,
            manual_validation=True,
        )
        root = ET.fromstring(str(what))
        form_params = {
            node.attrib['name']: node.attrib['value'] for node in root if node.attrib['type'] == 'hidden'
        }
        self.assertIn('PBX_AUTOSEULE', form_params)
        self.assertEqual(form_params['PBX_AUTOSEULE'], 'O')

    def test_invalid_signature(self):
        backend = eopayment.Payment('paybox', BACKEND_PARAMS)
        order_id = '20160216'
        transaction = '1234'
        reference = '%s!%s' % (transaction, order_id)
        data = {
            'montant': '4242',
            'reference': reference,
            'code_autorisation': 'A',
            'erreur': '00000',
            'date_transaction': '20200101',
            'heure_transaction': '01:01:01',
            'signature': 'a',
        }
        with pytest.raises(eopayment.ResponseError, match='invalid signature'):
            backend.response(urllib.urlencode(data))


@pytest.mark.parametrize(
    'name,value,result',
    [
        ('shared_secret', '1f', True),
        ('shared_secret', '1fxx', False),
        ('shared_secret', '1fa', False),
        ('shared_secret', '1fa2', True),
    ],
)
def test_param_validation(name, value, result):
    for param in paybox.Payment.description['parameters']:
        if param['name'] == name:
            assert param['validation'](value) is result
            break
    else:
        assert False, 'param %s not found' % name



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pytest

import eopayment


def test_build_request():
    backend = eopayment.Payment('sips2', {})
    transaction, f, form = backend.request(amount='12', last_name='Foo', first_name='Félix000000')
    data = [f for f in form.fields if f['name'] == 'Data']
    assert 'lix000000' not in data[0]['value']

    transaction, f, form = backend.request(amount='12')
    data = [f for f in form.fields if f['name'] == 'Data']
    assert 'statementReference=%s' % transaction in data[0]['value']

    transaction, f, form = backend.request(amount='12', info1='foobar')
    data = [f for f in form.fields if f['name'] == 'Data']
    assert 'statementReference=foobar' in data[0]['value']

    transaction, f, form = backend.request(amount='12', info1='foobar', capture_day='1')
    data = [f for f in form.fields if f['name'] == 'Data']
    assert 'captureDay=1' in data[0]['value']


def test_options():
    payment = eopayment.Payment('sips2', {'capture_mode': 'VALIDATION'})
    assert payment.backend.get_data()['captureMode'] == 'VALIDATION'

    payment = eopayment.Payment('sips2', {})
    assert 'captureDay' not in payment.backend.get_data()

    payment = eopayment.Payment('sips2', {'capture_day': '10'})
    assert 'captureDay' in payment.backend.get_data()


def test_parse_response():
    qs = '''Data=captureDay%3D0%7CcaptureMode%3DAUTHOR_CAPTURE%7CcurrencyCode%3D978%7CmerchantId%3D002001000000001%7CorderChannel%3DINTERNET%7CresponseCode%3D00%7CtransactionDateTime%3D2016-02-01T17%3A44%3A20%2B01%3A00%7CtransactionReference%3D668930%7CkeyVersion%3D1%7CacquirerResponseCode%3D00%7Camount%3D1200%7CauthorisationId%3D12345%7CcardCSCResultCode%3D4E%7CpanExpiryDate%3D201605%7CpaymentMeanBrand%3DMASTERCARD%7CpaymentMeanType%3DCARD%7CcustomerIpAddress%3D82.244.203.243%7CmaskedPan%3D5100%23%23%23%23%23%23%23%23%23%23%23%2300%7CorderId%3Dd4903de7027f4d56ac01634fd7ab9526%7CholderAuthentRelegation%3DN%7CholderAuthentStatus%3D3D_ERROR%7CtransactionOrigin%3DINTERNET%7CpaymentPattern%3DONE_SHOT&Seal=6ca3247765a19b45d25ad54ef4076483e7d55583166bd5ac9c64357aac097602&InterfaceVersion=HP_2.0&Encode='''  # noqa: E501
    backend = eopayment.Payment('sips2', {})
    response = backend.response(qs)
    assert response.signed
    assert response.transaction_date is None

    qs = '''Data=captureDay%3D0%7CcaptureMode%3DAUTHOR_CAPTURE%7CcurrencyCode%3D978%7CmerchantId%3D002001000000001%7CorderChannel%3DINTERNET%7CresponseCode%3D00%7CtransactionDateTime%3D2016-02-01T17%3A44%3A20%2B01%3A00%7CtransactionReference%3D668930%7CkeyVersion%3D1%7CacquirerResponseCode%3D00%7Camount%3D1200%7CauthorisationId%3D12345%7CcardCSCResultCode%3D4E%7CpanExpiryDate%3D201605%7CpaymentMeanBrand%3DMASTERCARD%7CpaymentMeanType%3DCARD%7CcustomerIpAddress%3D82.244.203.243%7CmaskedPan%3D5100%23%23%23%23%23%23%23%23%23%23%23%2300%7CorderId%3Dd4903de7027f4d56ac01634fd7ab9526%7CholderAuthentRelegation%3DN%7CholderAuthentStatus%3D3D_ERROR%7CtransactionOrigin%3DINTERNET%7CpaymentPattern%3DONE_SHOT%7CtransactionDateTime%3D2020-01-01%2001:01:01&Seal=6ca3247765a19b45d25ad54ef4076483e7d55583166bd5ac9c64357aac097602&InterfaceVersion=HP_2.0&Encode='''  # noqa: E501
    response = backend.response(qs)
    assert not response.signed
    assert response.transaction_date.isoformat() == '2020-01-01T01:01:01+01:00'

    with pytest.raises(eopayment.ResponseError, match='missing Data, Seal or InterfaceVersion'):
        backend.response('foo=bar')



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import uuid
import warnings

from six.moves.urllib.parse import parse_qs, urlencode

from .common import ERROR, PAID, URL, WAITING, PaymentCommon, PaymentResponse, ResponseError, _, force_text

__all__ = ['Payment']


SERVICE_URL = 'http://dummy-payment.demo.entrouvert.com/'
LOGGER = logging.getLogger(__name__)


class Payment(PaymentCommon):
    """
    Dummy implementation of the payment interface.

    It is used with a dummy implementation of a bank payment service that
    you can find on:

        http://dummy-payment.demo.entrouvert.com/

    You must pass the following keys inside the options dictionnary:
     - dummy_service_url, the URL of the dummy payment service, it defaults
       to the one operated by Entr'ouvert.
     - automatic_return_url: where to POST to notify the service of a
       payment
     - origin: a human string to display to the user about the origin of
       the request.
     - siret: an identifier for the eCommerce site, fake.
     - normal_return_url: the return URL for the user (can be overriden on a
       per request basis).
    """

    description = {
        'caption': 'Dummy payment backend',
        'parameters': [
            {
                'name': 'normal_return_url',
                'caption': _('Normal return URL'),
                'default': '',
                'required': True,
            },
            {
                'name': 'automatic_return_url',
                'caption': _('Automatic return URL'),
                'required': False,
            },
            {
                'name': 'dummy_service_url',
                'caption': _('URL of the dummy payment service'),
                'default': SERVICE_URL,
                'type': str,
            },
            {
                'name': 'origin',
                'caption': _('name of the requesting service, ' 'to present in the user interface'),
                'type': str,
                'default': 'origin',
            },
            {
                'name': 'consider_all_response_signed',
                'caption': _(
                    'All response will be considered as signed '
                    '(to test payment locally for example, as you '
                    'cannot received the signed callback)'
                ),
                'type': bool,
                'default': False,
            },
            {
                'name': 'number',
                'caption': 'dummy integer input test',
                'type': int,
            },
            {
                'name': 'choice',
                'caption': 'dummy choice input test',
                'choices': ['a', 'b'],
            },
            {
                'name': 'choices',
                'caption': 'dummy choices input test',
                'choices': ['a', 'b'],
                'type': list,
            },
            {
                'name': 'direct_notification_url',
                'caption': _('direct notification url (replaced by automatic_return_url)'),
                'type': str,
                'deprecated': True,
            },
            {
                'name': 'next_url (replaced by normal_return_url)',
                'caption': _('Return URL for the user'),
                'type': str,
                'deprecated': True,
            },
            {
                'name': 'capture_day',
                'type': str,
            },
        ],
    }

    def request(
        self,
        amount,
        name=None,
        address=None,
        email=None,
        phone=None,
        orderid=None,
        info1=None,
        info2=None,
        info3=None,
        next_url=None,
        capture_day=None,
        subject=None,
        **kwargs,
    ):
        self.logger.debug(
            '%s amount %s name %s address %s email %s phone %s'
            ' next_url %s info1 %s info2 %s info3 %s kwargs: %s',
            __name__,
            amount,
            name,
            address,
            email,
            phone,
            info1,
            info2,
            info3,
            next_url,
            kwargs,
        )
        transaction_id = str(uuid.uuid4().hex)
        normal_return_url = self.normal_return_url
        if next_url and not normal_return_url:
            warnings.warn(
                "passing next_url to request() is deprecated, " "set normal_return_url in options",
                DeprecationWarning,
            )
            normal_return_url = next_url
        automatic_return_url = self.automatic_return_url
        if self.direct_notification_url and not automatic_return_url:
            warnings.warn(
                "direct_notification_url option is deprecated, " "use automatic_return_url",
                DeprecationWarning,
            )
            automatic_return_url = self.direct_notification_url
        query = {
            'transaction_id': transaction_id,
            'amount': amount,
            'email': email,
            'return_url': normal_return_url or '',
            'direct_notification_url': automatic_return_url or '',
            'origin': self.origin,
        }
        query.update(
            dict(
                name=name,
                address=address,
                email=email,
                phone=phone,
                orderid=orderid,
                info1=info1,
                info2=info2,
                info3=info3,
            )
        )
        if capture_day is not None:
            query['capture_day'] = str(capture_day)
        if subject is not None:
            query['subject'] = subject
        for key in list(query.keys()):
            if query[key] is None:
                del query[key]
        url = '%s?%s' % (SERVICE_URL, urlencode(query))
        return transaction_id, URL, url

    def response(self, query_string, logger=LOGGER, **kwargs):
        form = parse_qs(force_text(query_string))
        if 'transaction_id' not in form:
            raise ResponseError('missing transaction_id')
        transaction_id = form.get('transaction_id', [''])[0]
        form[self.BANK_ID] = transaction_id

        signed = 'signed' in form
        if signed:
            content = 'signature ok'
        else:
            content = None
        signed = signed or self.consider_all_response_signed
        result = PAID if 'ok' in form else ERROR
        if 'waiting' in form:
            result = WAITING

        response = PaymentResponse(
            result=result,
            signed=signed,
            bank_data=form,
            return_content=content,
            order_id=transaction_id,
            transaction_id=transaction_id,
            bank_status=form.get('reason'),
            test=True,
        )
        return response

    def validate(self, amount, bank_data, **kwargs):
        return {}

    def cancel(self, amount, bank_data, **kwargs):
        return {}



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import datetime as dt
import hashlib
import hmac
import random
import re
import string
import warnings

import pytz
from six.moves.urllib import parse as urlparse

from .cb import translate_cb_error_code
from .common import (
    CANCELLED,
    DENIED,
    ERROR,
    FORM,
    PAID,
    Form,
    PaymentCommon,
    PaymentResponse,
    ResponseError,
    _,
    force_byte,
    force_text,
)

__all__ = ['Payment']

VADS_TRANS_DATE = 'vads_trans_date'
VADS_AUTH_NUMBER = 'vads_auth_number'
VADS_AUTH_RESULT = 'vads_auth_result'
VADS_RESULT = 'vads_result'
VADS_EXTRA_RESULT = 'vads_extra_result'
VADS_CUST_EMAIL = 'vads_cust_email'
VADS_CUST_NAME = 'vads_cust_name'
VADS_CUST_PHONE = 'vads_cust_phone'
VADS_CUST_INFO1 = 'vads_order_info'
VADS_CUST_INFO2 = 'vads_order_info2'
VADS_CUST_INFO3 = 'vads_order_info3'
VADS_CUST_FIRST_NAME = 'vads_cust_first_name'
VADS_CUST_LAST_NAME = 'vads_cust_last_name'
VADS_URL_RETURN = 'vads_url_return'
VADS_AMOUNT = 'vads_amount'
VADS_SITE_ID = 'vads_site_id'
VADS_TRANS_ID = 'vads_trans_id'
SIGNATURE = 'signature'
VADS_CTX_MODE = 'vads_ctx_mode'
VADS_EFFECTIVE_CREATION_DATE = 'vads_effective_creation_date'
VADS_EOPAYMENT_TRANS_ID = 'vads_ext_info_eopayment_trans_id'


def isonow():
    return dt.datetime.utcnow().isoformat('T').replace('-', '').replace('T', '').replace(':', '')[:14]


def parse_utc(value):
    try:
        naive_dt = dt.datetime.strptime(value, '%Y%m%d%H%M%S')
    except ValueError:
        return None
    return pytz.utc.localize(naive_dt)


class Parameter:
    def __init__(
        self,
        name,
        ptype,
        code,
        max_length=None,
        length=None,
        needed=False,
        default=None,
        choices=None,
        description=None,
        help_text=None,
        scope='global',
    ):
        self.name = name
        self.ptype = ptype
        self.code = code
        self.max_length = max_length
        self.length = length
        self.needed = needed
        self.default = default
        self.choices = choices
        self.description = description
        self.help_text = help_text
        self.scope = scope

    def check_value(self, value):
        if self.length and len(value) != self.length:
            return False
        if self.max_length and len(value) > self.max_length:
            return False
        if self.choices and value not in self.choices:
            return False
        if value == '':
            return True
        value = value.replace('.', '')
        if self.ptype == 'n':
            return value.isdigit()
        elif self.ptype == 'an':
            return value.isalnum()
        elif self.ptype == 'an-':
            return value.replace('-', '').isalnum()
        elif self.ptype == 'an;':
            return value.replace(';', '').isalnum()
        return True


PARAMETERS = [
    # amount as euro cents
    Parameter(
        'vads_action_mode', None, 47, needed=True, default='INTERACTIVE', choices=('SILENT', 'INTERACTIVE')
    ),
    Parameter('vads_amount', 'n', 9, max_length=12, needed=True),
    Parameter('vads_capture_delay', 'n', 6, max_length=3, default=''),
    # Same as 'vads_capture_delay' but matches other backend naming for
    # deferred payment
    Parameter('capture_day', 'n', 6, max_length=3, default=''),
    Parameter('vads_contrib', 'ans', 31, max_length=255, default='eopayment'),
    # defaut currency = EURO, norme ISO4217
    Parameter('vads_currency', 'n', 10, length=3, default='978', needed=True),
    Parameter('vads_cust_address', 'an', 19, max_length=255),
    # code ISO 3166
    Parameter('vads_cust_country', 'a', 22, length=2, default='FR'),
    Parameter('vads_cust_email', 'an@', 15, max_length=127),
    Parameter('vads_cust_id', 'an', 16, max_length=63),
    Parameter('vads_cust_name', 'ans', 18, max_length=127),
    Parameter('vads_cust_phone', 'an', 23, max_length=63),
    Parameter('vads_cust_title', 'an', 17, max_length=63),
    Parameter('vads_cust_city', 'an', 21, max_length=63),
    Parameter('vads_cust_zip', 'an', 20, max_length=63),
    Parameter('vads_ctx_mode', 'a', 11, needed=True, choices=('TEST', 'PRODUCTION'), default='TEST'),
    # ISO 639 code
    Parameter('vads_language', 'a', 12, length=2, default='fr'),
    Parameter('vads_order_id', 'an-', 13, max_length=32),
    Parameter(
        'vads_order_info',
        'an',
        14,
        max_length=255,
        description="Complément d'information 1",
        scope='transaction',
    ),
    Parameter(
        'vads_order_info2',
        'an',
        14,
        max_length=255,
        description="Complément d'information 2",
        scope='transaction',
    ),
    Parameter(
        'vads_order_info3',
        'an',
        14,
        max_length=255,
        description="Complément d'information 3",
        scope='transaction',
    ),
    Parameter('vads_page_action', None, 46, needed=True, default='PAYMENT', choices=('PAYMENT',)),
    Parameter(
        'vads_payment_cards',
        'an;',
        8,
        max_length=127,
        default='',
        description='Liste des cartes de paiement acceptées',
        help_text='vide ou des valeurs sépareés par un point-virgule '
        'parmi AMEX, AURORE-MULTI, BUYSTER, CB, COFINOGA, '
        'E-CARTEBLEUE, MASTERCARD, JCB, MAESTRO, ONEY, '
        'ONEY_SANDBOX, PAYPAL, PAYPAL_SB, PAYSAFECARD, '
        'VISA',
        scope='transaction',
    ),
    # must be SINGLE or MULTI with parameters
    Parameter('vads_payment_config', '', 7, default='SINGLE', choices=('SINGLE', 'MULTI'), needed=True),
    Parameter('vads_return_mode', None, 48, default='GET', choices=('', 'NONE', 'POST', 'GET')),
    Parameter('signature', 'an', None, length=40),
    Parameter('vads_site_id', 'n', 2, length=8, needed=True, description='Identifiant de la boutique'),
    Parameter('vads_theme_config', 'ans', 32, max_length=255),
    Parameter(VADS_TRANS_DATE, 'n', 4, length=14, needed=True, default=isonow),
    # https://paiement.systempay.fr/doc/fr-FR/form-payment/reference/vads-trans-id.html
    Parameter('vads_trans_id', 'an', 3, length=6, needed=True),
    Parameter('vads_validation_mode', 'n', 5, max_length=1, choices=('', '0', '1'), default=''),
    Parameter('vads_version', 'an', 1, default='V2', needed=True, choices=('V2',)),
    Parameter('vads_url_success', 'ans', 24, max_length=1024),
    Parameter('vads_url_referral', 'ans', 26, max_length=127),
    Parameter('vads_url_refused', 'ans', 25, max_length=1024),
    Parameter('vads_url_cancel', 'ans', 27, max_length=1024),
    Parameter('vads_url_error', 'ans', 29, max_length=1024),
    Parameter('vads_url_return', 'ans', 28, max_length=1024),
    Parameter('vads_user_info', 'ans', 61, max_length=255),
    Parameter('vads_contracts', 'ans', 62, max_length=255),
    Parameter(VADS_CUST_FIRST_NAME, 'ans', 104, max_length=63),
    Parameter(VADS_CUST_LAST_NAME, 'ans', 104, max_length=63),
]
PARAMETER_MAP = {parameter.name: parameter for parameter in PARAMETERS}


def add_vads(kwargs):
    new_vargs = {}
    for k, v in kwargs.items():
        if k.startswith('vads_'):
            new_vargs[k] = v
        else:
            new_vargs['vads_' + k] = v
    return new_vargs


def check_vads(kwargs, exclude=[]):
    for parameter in PARAMETERS:
        name = parameter.name
        if name not in kwargs and name not in exclude and parameter.needed:
            raise ValueError('parameter %s must be defined' % name)
        if name in kwargs and not parameter.check_value(kwargs[name]):
            raise ValueError(
                'parameter %s value %s is not of the type %s' % (name, kwargs[name], parameter.ptype)
            )


class Payment(PaymentCommon):
    """
    Produce request for and verify response from the SystemPay payment
    gateway.

        >>> gw =Payment(dict(secret_test='xxx', secret_production='yyyy',
                             site_id=123, ctx_mode='PRODUCTION'))
        >>> print gw.request(100)
        ('20120525093304_188620',
        'https://paiement.systempay.fr/vads-payment/?vads_url_return=http%3A%2F%2Furl.de.retour%2Fretour.php&vads_cust_country=FR&vads_site_id=&vads_payment_config=SINGLE&vads_trans_id=188620&vads_action_mode=INTERACTIVE&vads_contrib=eopayment&vads_page_action=PAYMENT&vads_trans_date=20120525093304&vads_ctx_mode=TEST&vads_validation_mode=&vads_version=V2&vads_payment_cards=&signature=5d412498ab523627ec5730a09118f75afa602af5&vads_language=fr&vads_capture_delay=&vads_currency=978&vads_amount=100&vads_return_mode=NONE',
        {'vads_url_return': 'http://url.de.retour/retour.php',
        'vads_cust_country': 'FR', 'vads_site_id': '',
        'vads_payment_config': 'SINGLE', 'vads_trans_id': '188620',
        'vads_action_mode': 'INTERACTIVE', 'vads_contrib': 'eopayment',
        'vads_page_action': 'PAYMENT', 'vads_trans_date': '20120525093304',
        'vads_ctx_mode': 'TEST', 'vads_validation_mode': '',
        'vads_version': 'V2', 'vads_payment_cards': '', 'signature':
        '5d412498ab523627ec5730a09118f75afa602af5', 'vads_language': 'fr',
        'vads_capture_delay': '', 'vads_currency': '978', 'vads_amount': 100,
        'vads_return_mode': 'NONE'})

    """

    has_free_transaction_id = True
    service_url = "https://paiement.systempay.fr/vads-payment/"
    signature_algo = 'sha1'

    description = {
        'caption': 'SystemPay, système de paiement du groupe BPCE',
        'parameters': [
            {
                'name': 'normal_return_url',
                'caption': _('Normal return URL'),
                'default': '',
                'required': True,
            },
            {
                'name': 'automatic_return_url',
                'caption': _('Automatic return URL (ignored, must be set in Payzen/SystemPay backoffice)'),
                'required': False,
            },
            {
                'name': 'service_url',
                'default': service_url,
                'caption': 'URL du service de paiement',
                'help_text': 'ne pas modifier si vous ne savez pas',
                'validation': lambda x: x.startswith('http'),
                'required': True,
            },
            {
                'name': 'secret_test',
                'caption': 'Secret pour la configuration de TEST',
                'validation': lambda value: str.isalnum(value),
                'required': True,
            },
            {
                'name': 'secret_production',
                'caption': 'Secret pour la configuration de PRODUCTION',
                'validation': lambda value: str.isalnum(value),
            },
            {
                'name': 'signature_algo',
                'caption': 'Algorithme de signature',
                'default': 'sha1',
                'choices': (
                    ('sha1', 'SHA-1'),
                    ('hmac_sha256', 'HMAC-SHA-256'),
                ),
            },
            {
                'name': 'manual_validation',
                'caption': 'Validation manuelle',
                'type': bool,
                'default': False,
                'scope': 'transaction',
            },
        ],
    }

    for name in (
        'vads_ctx_mode',
        VADS_SITE_ID,
        'vads_order_info',
        'vads_order_info2',
        'vads_order_info3',
        'vads_payment_cards',
        'vads_payment_config',
        'capture_day',
    ):
        parameter = PARAMETER_MAP[name]

        def check_value(parameter):
            def validate(value):
                return parameter.check_value(value)

            return validate

        x = {
            'name': name,
            'caption': parameter.description or name,
            'validation': check_value(parameter),
            'default': parameter.default,
            'required': parameter.needed,
            'help_text': parameter.help_text,
            'max_length': parameter.max_length,
            'scope': parameter.scope,
        }
        description['parameters'].append(x)

    def __init__(self, options, logger=None):
        super().__init__(options, logger=logger)
        options = add_vads(options)
        self.options = options

    def make_vads_trans_id(self):
        # vads_trans_id must be 6 alphanumeric characters,
        # trans_id starting with 9 are reserved for the systempay backoffice
        # https://paiement.systempay.fr/doc/fr-FR/form-payment/reference/vads-trans-id.html
        gen = random.SystemRandom()

        alphabet = string.ascii_letters + string.digits
        first_letter_alphabet = alphabet.replace('9', '')
        vads_trans_id = gen.choice(first_letter_alphabet) + ''.join(gen.choice(alphabet) for i in range(5))
        return vads_trans_id

    def request(
        self,
        amount,
        name=None,
        first_name=None,
        last_name=None,
        address=None,
        email=None,
        phone=None,
        orderid=None,
        info1=None,
        info2=None,
        info3=None,
        next_url=None,
        manual_validation=None,
        transaction_id=None,
        **kwargs,
    ):
        """
        Create the URL string to send a request to SystemPay
        """
        self.logger.debug(
            '%s amount %s name %s address %s email %s phone %s '
            'next_url %s info1 %s info2 %s info3 %s kwargs: %s',
            __name__,
            amount,
            name,
            address,
            email,
            phone,
            info1,
            info2,
            info3,
            next_url,
            kwargs,
        )
        # amount unit is cents
        amount = '%.0f' % (100 * amount)
        kwargs.update(add_vads({'amount': force_text(amount)}))
        if int(amount) < 0:
            raise ValueError('amount must be an integer >= 0')
        normal_return_url = self.normal_return_url
        if next_url:
            warnings.warn(
                "passing next_url to request() is deprecated, " "set normal_return_url in options",
                DeprecationWarning,
            )
            normal_return_url = next_url
        if normal_return_url:
            kwargs[VADS_URL_RETURN] = force_text(normal_return_url)
        if name is not None:
            kwargs['vads_cust_name'] = force_text(name)
        if first_name is not None:
            kwargs[VADS_CUST_FIRST_NAME] = force_text(first_name)
        if last_name is not None:
            kwargs[VADS_CUST_LAST_NAME] = force_text(last_name)

        if address is not None:
            kwargs['vads_cust_address'] = force_text(address)
        if email is not None:
            kwargs['vads_cust_email'] = force_text(email)
        if phone is not None:
            kwargs['vads_cust_phone'] = force_text(phone)
        if info1 is not None:
            kwargs['vads_order_info'] = force_text(info1)
        if info2 is not None:
            kwargs['vads_order_info2'] = force_text(info2)
        if info3 is not None:
            kwargs['vads_order_info3'] = force_text(info3)
        if orderid is not None:
            # check orderid format first
            name = 'vads_order_id'
            orderid = force_text(orderid)
            ptype = 'an-'
            p = Parameter(name, ptype, 13, max_length=32)
            if not p.check_value(orderid):
                raise ValueError('%s value %s is not of the type %s' % (name, orderid, ptype))
            kwargs[name] = orderid

        vads_trans_id = self.make_vads_trans_id()
        assert re.match(r'^[0-9a-zA-Z]{6}$', vads_trans_id)

        kwargs[VADS_TRANS_ID] = vads_trans_id
        fields = kwargs
        for parameter in PARAMETERS:
            name = parameter.name
            # import default parameters from configuration
            if name not in fields and name in self.options:
                fields[name] = force_text(self.options[name])
            # import default parameters from module
            if name not in fields and parameter.default is not None:
                if callable(parameter.default):
                    fields[name] = parameter.default()
                else:
                    fields[name] = parameter.default
        capture_day = fields.pop('capture_day')
        if capture_day:
            fields['vads_capture_delay'] = capture_day
        if manual_validation:
            fields['vads_validation_mode'] = '1'
        check_vads(fields)
        if transaction_id:
            fields[VADS_EOPAYMENT_TRANS_ID] = transaction_id
        else:
            transaction_id = '%s_%s' % (fields[VADS_TRANS_DATE], vads_trans_id)
        fields[SIGNATURE] = force_text(self.signature(fields))
        self.logger.debug('%s request contains fields: %s', __name__, fields)
        self.logger.debug('%s transaction id: %s', __name__, transaction_id)
        form = Form(
            url=self.service_url,
            method='POST',
            fields=[
                {
                    'type': 'hidden',
                    'name': force_text(field_name),
                    'value': force_text(field_value),
                }
                for field_name, field_value in fields.items()
            ],
        )
        return transaction_id, FORM, form

    RESULT_MAP = {
        '00': {'message': 'Paiement réalisé avec succés.', 'result': PAID},
        '02': {'message': 'Le commerçant doit contacter la banque du porteur.'},
        '05': {'message': 'Paiement refusé.', 'result': DENIED},
        '17': {'message': 'Annulation client.', 'result': CANCELLED},
        '30': {'message': 'Erreur de format.'},
        '96': {'message': 'Erreur technique lors du paiement.'},
    }

    EXTRA_RESULT_MAP = {
        '': {'message': 'Pas de contrôle effectué.'},
        '00': {'message': 'Tous les contrôles se sont déroulés avec succés.'},
        '02': {'message': 'La carte a dépassé l\'encours autorisé.'},
        '03': {'message': 'La carte appartient à la liste grise du commerçant.'},
        '04': {
            'messaǵe': 'Le pays d\'émission de la carte appartient à la liste grise du '
            'commerçant ou le pays d\'émission de la carte n\'appartient pas à la '
            'liste blanche du commerçant.'
        },
        '05': {'message': 'L’adresse IP appartient à la liste grise du marchand.'},
        '06': {'message': 'Le code bin appartient à la liste grise du marchand.'},
        '07': {'message': 'Détection d’une e-carte bleue.'},
        '08': {'message': 'Détection d’une carte commerciale nationale.'},
        '09': {'message': 'Détection d’une carte commerciale étrangère.'},
        '14': {'message': 'Détection d’une carte à autorisation systématique.'},
        '30': {'message': 'Le pays de l’adresse IP appartient à la liste grise.'},
        '99': {
            'message': 'Problème technique recontré par le serveur lors du traitement '
            'd\'un des contrôles locauxi.'
        },
    }

    @classmethod
    def make_eopayment_result(cls, fields):
        # https://paiement.systempay.fr/doc/fr-FR/payment-file/oneclick-payment/vads-result.html
        # https://paiement.systempay.fr/doc/fr-FR/payment-file/oneclick-payment/vads-auth-result.html
        # https://paiement.systempay.fr/doc/fr-FR/payment-file/oneclick-payment/vads-extra-result.html
        vads_result = fields.get(VADS_RESULT)
        vads_auth_result = fields.get(VADS_AUTH_RESULT)
        vads_extra_result = fields.get(VADS_EXTRA_RESULT)

        # map to human messages and update return
        vads_result_message = cls.RESULT_MAP.get(vads_result, {}).get('message')
        if vads_result_message:
            fields[VADS_RESULT + '_message'] = vads_result_message

        vads_extra_result_message = cls.EXTRA_RESULT_MAP.get(vads_extra_result, {}).get('message')
        if vads_extra_result_message:
            fields[VADS_EXTRA_RESULT + '_message'] = vads_extra_result_message

        vads_auth_result_message, auth_eopayment_result = translate_cb_error_code(vads_auth_result)
        if vads_auth_result_message:
            fields[VADS_AUTH_RESULT + '_message'] = vads_auth_result_message

        # now build eopayment resume
        if vads_result is None:
            return ERROR, 'absence de champ vads_result'
        if vads_result_message is None:
            return ERROR, 'valeur vads_result inconnue'

        result = cls.RESULT_MAP[vads_result].get('result', ERROR)
        message = vads_result_message
        if vads_auth_result_message and (vads_result != '00' or vads_result != vads_auth_result):
            message += ' ' + vads_auth_result_message
        if vads_result in ('00', '05', '30') and vads_extra_result_message and vads_extra_result != '':
            message += ' ' + vads_extra_result_message
        if result == ERROR and auth_eopayment_result not in (PAID, ERROR, None):
            result = auth_eopayment_result
        return result, message

    def response(self, query_string, **kwargs):
        fields = urlparse.parse_qs(query_string, True)
        if not set(fields) >= {SIGNATURE, VADS_CTX_MODE, VADS_AUTH_RESULT}:
            raise ResponseError('missing %s, %s or %s' % (SIGNATURE, VADS_CTX_MODE, VADS_AUTH_RESULT))
        for key, value in fields.items():
            fields[key] = value[0]
        copy = fields.copy()
        signature = self.signature(fields)
        result, message = self.make_eopayment_result(copy)
        self.logger.debug('checking systempay response on: %r', copy)
        signature_result = signature == fields[SIGNATURE]
        if not signature_result:
            self.logger.debug('signature check: %s <!> %s', signature, fields[SIGNATURE])

        if not signature_result:
            message += ' signature invalide.'

        test = fields[VADS_CTX_MODE] == 'TEST'
        vads_eopayment_trans_id = fields.get(VADS_EOPAYMENT_TRANS_ID)
        vads_trans_date = fields.get(VADS_TRANS_DATE)
        vads_trans_id = fields.get(VADS_TRANS_ID)
        if vads_eopayment_trans_id:
            transaction_id = vads_eopayment_trans_id
        elif vads_trans_date and vads_trans_id:
            transaction_id = vads_trans_date + '_' + vads_trans_id
        else:
            raise ResponseError('backend error', message)
        # the VADS_AUTH_NUMBER is the number to match payment in bank logs
        copy[self.BANK_ID] = copy.get(VADS_AUTH_NUMBER, '')
        transaction_date = None
        if VADS_EFFECTIVE_CREATION_DATE in fields:
            transaction_date = parse_utc(fields[VADS_EFFECTIVE_CREATION_DATE])
        response = PaymentResponse(
            result=result,
            signed=signature_result,
            bank_data=copy,
            order_id=transaction_id,
            transaction_id=transaction_id,
            bank_status=message,
            transaction_date=transaction_date,
            test=test,
        )
        return response

    def sha1_sign(self, secret, signed_data):
        return hashlib.sha1(signed_data).hexdigest()

    def hmac_sha256_sign(self, secret, signed_data):
        digest = hmac.HMAC(secret, digestmod=hashlib.sha256, msg=signed_data).digest()
        return base64.b64encode(digest)

    def signature(self, fields):
        self.logger.debug('got fields %s to sign' % fields)
        ordered_keys = sorted(key for key in fields.keys() if key.startswith('vads_'))
        self.logger.debug('ordered keys %s' % ordered_keys)
        ordered_fields = [force_byte(fields[key]) for key in ordered_keys]
        secret = force_byte(getattr(self, 'secret_%s' % fields['vads_ctx_mode'].lower()))
        signed_data = b'+'.join(ordered_fields)
        signed_data = b'%s+%s' % (signed_data, secret)
        self.logger.debug('generating signature on «%s»', signed_data)
        sign_method = getattr(self, '%s_sign' % self.signature_algo)
        sign = sign_method(secret, signed_data)
        self.logger.debug('signature «%s»', sign)
        return force_text(sign)

    @classmethod
    def guess(self, *, method=None, query_string=None, body=None, headers=None, backends=(), **kwargs):
        for content in [query_string, body]:
            if isinstance(content, bytes):
                try:
                    content = content.decode()
                except UnicodeDecodeError:
                    pass
            if isinstance(content, str):
                fields = urlparse.parse_qs(content)
                if not set(fields) >= {SIGNATURE, VADS_CTX_MODE, VADS_AUTH_RESULT}:
                    continue
                vads_eopayment_trans_id = fields.get(VADS_EOPAYMENT_TRANS_ID)
                vads_trans_date = fields.get(VADS_TRANS_DATE)
                vads_trans_id = fields.get(VADS_TRANS_ID)
                if vads_eopayment_trans_id:
                    return vads_eopayment_trans_id[0]
                elif vads_trans_date and vads_trans_id:
                    return vads_trans_date[0] + '_' + vads_trans_id[0]
        return None



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import requests
from six.moves.urllib.parse import parse_qs, urljoin

from .common import (
    CANCELLED,
    ERROR,
    PAID,
    URL,
    WAITING,
    PaymentCommon,
    PaymentException,
    PaymentResponse,
    ResponseError,
    _,
)

__all__ = ['Payment']


class Payment(PaymentCommon):
    '''Implements EMS API, see https://dev.online.emspay.eu/.'''

    service_url = 'https://api.online.emspay.eu/v1/'

    description = {
        'caption': _('Keyware payment backend'),
        'parameters': [
            {
                'name': 'normal_return_url',
                'caption': _('Normal return URL'),
                'required': True,
            },
            {
                'name': 'automatic_return_url',
                'caption': _('Asychronous return URL'),
                'required': True,
            },
            {
                'name': 'service_url',
                'caption': _('URL of the payment service'),
                'default': service_url,
                'type': str,
                'validation': lambda x: x.startswith('https'),
            },
            {
                'name': 'api_key',
                'caption': _('API key'),
                'required': True,
            },
        ],
    }

    def request(self, amount, email=None, first_name=None, last_name=None, **kwargs):
        amount = int(self.clean_amount(amount, min_amount=0.01))

        body = {
            'currency': 'EUR',
            'amount': amount,
            'return_url': self.normal_return_url,
            'webhook_url': self.automatic_return_url,
            'customer': {
                'email_address': email,
                'first_name': first_name,
                'last_name': last_name,
            },
        }

        resp = self.call_endpoint('POST', 'orders', data=body)
        return resp['id'], URL, resp['order_url']

    def response(self, query_string, **kwargs):
        fields = parse_qs(query_string, keep_blank_values=True)
        order_id = fields['order_id'][0]
        resp = self.call_endpoint('GET', 'orders/' + order_id)

        # XXX: to add accepted we need to handle the capture mode (manual or
        # delayed), see
        # https://dev.online.emspay.eu/rest-api/features/authorizations-captures-and-voiding
        status = resp['status']
        if status == 'completed':
            result = PAID
        elif status in ('new', 'processing'):
            result = WAITING
        elif status in ('cancelled', 'expired'):
            result = CANCELLED
        else:
            result = ERROR

        response = PaymentResponse(
            result=result,
            signed=True,
            bank_data=resp,
            order_id=order_id,
            transaction_id=order_id,
            bank_status=status,
            test=bool('is-test' in resp.get('flags', [])),
        )
        return response

    def cancel(self, amount, bank_data, **kwargs):
        order_id = bank_data['id']
        resp = self.call_endpoint('DELETE', 'orders/' + order_id)
        status = resp['status']
        if not status == 'cancelled':
            raise ResponseError('expected "cancelled" status, got "%s" instead' % status)
        return resp

    def call_endpoint(self, method, endpoint, data=None):
        url = urljoin(self.service_url, endpoint)
        try:
            response = requests.request(method, url, auth=(self.api_key, ''), json=data)
        except requests.exceptions.RequestException as e:
            raise PaymentException('%s error on endpoint "%s": %s' % (method, endpoint, e))
        try:
            result = response.json()
        except ValueError:
            self.logger.debug('received invalid json %r', response.text)
            raise PaymentException(
                '%s on endpoint "%s" returned invalid JSON: %s' % (method, endpoint, response.text)
            )
        self.logger.debug('received "%s" with status %s', result, response.status_code)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise PaymentException('%s error on endpoint "%s": %s "%s"' % (method, endpoint, e, result))
        return result



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from datetime import datetime, timedelta

import pytest
from six.moves.urllib import parse as urlparse

import eopayment
from eopayment import ResponseError
from eopayment.systempayv2 import PAID, VADS_CUST_FIRST_NAME, VADS_CUST_LAST_NAME, Payment

PARAMS = {
    'secret_test': '1122334455667788',
    'vads_site_id': '12345678',
    'vads_ctx_mode': 'TEST',
    'vads_trans_date': '20090501193530',
    'signature_algo': 'sha1',
}


@pytest.fixture
def backend():
    return eopayment.Payment('systempayv2', PARAMS)


def get_field(form, field_name):
    for field in form.fields:
        if field['name'] == field_name:
            return field


def test_systempayv2(caplog):
    caplog.set_level(0)
    p = Payment(PARAMS)
    data = {'amount': 15.24, 'orderid': '654321', 'first_name': 'Jean Michél', 'last_name': 'Mihaï'}
    qs = (
        'vads_version=V2&vads_page_action=PAYMENT&vads_action_mode=INTERACTIV'
        'E&vads_payment_config=SINGLE&vads_site_id=12345678&vads_ctx_mode=TES'
        'T&vads_trans_id=654321&vads_trans_date=20090501193530&vads_amount=15'
        '24&vads_currency=978&vads_cust_first_name=Jean+Mich%C3%A9l&vads_cust_last_name=Mihaï'
    )
    qs = urlparse.parse_qs(qs)
    for key in qs.keys():
        qs[key] = qs[key][0]
    assert p.signature(qs) == 'cf1ccac335590a33d6c243345c4f4531a0316b7f'
    transaction_id, f, form = p.request(**data)

    # check that user first and last names are unicode
    for field in form.fields:
        if field['name'] in (VADS_CUST_FIRST_NAME, VADS_CUST_LAST_NAME):
            assert field['value'] in ('Jean Michél', 'Mihaï')

    response_qs = (
        'vads_amount=1042&vads_auth_mode=FULL&vads_auth_number=3feadf'
        '&vads_auth_result=00&vads_capture_delay=0&vads_card_brand=CB'
        '&vads_result=00'
        '&vads_card_number=497010XXXXXX0000'
        '&vads_payment_certificate=582ba2b725057618706d7a06e9e59acdbf69ff53'
        '&vads_ctx_mode=TEST&vads_currency=978&vads_effective_amount=1042'
        '&vads_site_id=70168983&vads_trans_date=20161013101355'
        '&vads_trans_id=226787&vads_trans_uuid=4b5053b3b1fe4b02a07753e7a'
        '&vads_effective_creation_date=20200330162530'
        '&signature=c17fab393f94dc027dc029510c85d5fc46c4710f'
    )
    response = p.response(response_qs)
    assert response.result == PAID
    assert response.signed
    assert response.transaction_date
    assert response.transaction_date.isoformat() == '2020-03-30T16:25:30+00:00'

    PARAMS['signature_algo'] = 'hmac_sha256'
    p = Payment(PARAMS)
    assert p.signature(qs) == 'aHrJ7IzSGFa4pcYA8kh99+M/xBzoQ4Odnu3f4BUrpIA='
    response_qs = (
        'vads_amount=1042&vads_auth_mode=FULL&vads_auth_number=3feadf'
        '&vads_result=00'
        '&vads_auth_result=00&vads_capture_delay=0&vads_card_brand=CB'
        '&vads_card_number=497010XXXXXX0000'
        '&vads_payment_certificate=582ba2b725057618706d7a06e9e59acdbf69ff53'
        '&vads_ctx_mode=TEST&vads_currency=978&vads_effective_amount=1042'
        '&vads_site_id=70168983&vads_trans_date=20161013101355'
        '&vads_trans_id=226787&vads_trans_uuid=4b5053b3b1fe4b02a07753e7a'
        '&vads_effective_creation_date=20200330162530'
        '&signature=Wbz3bP6R6wDvAwb2HnSiH9%2FiUUoRVCxK7mdLtCMz8Xw%3D'
    )
    response = p.response(response_qs)
    assert response.result == PAID
    assert response.signed

    # bad response
    with pytest.raises(ResponseError, match='missing signature, vads_ctx_mode or vads_auth_result'):
        p.response('foo=bar')


def test_systempayv2_deferred_payment():
    default_params = {
        'secret_test': '1122334455667788',
        'vads_site_id': '12345678',
        'vads_ctx_mode': 'TEST',
    }
    default_data = {'amount': 15.24, 'orderid': '654321', 'first_name': 'John', 'last_name': 'Doe'}

    # default vads_capture_delay used
    params = default_params.copy()
    params['vads_capture_delay'] = 1

    backend = eopayment.Payment('systempayv2', params)
    data = default_data.copy()
    transaction_id, f, form = backend.request(**data)
    assert get_field(form, 'vads_capture_delay')['value'] == '1'

    # vads_capture_delay can used in request and
    # override default vads_capture_delay
    params = default_params.copy()
    params['vads_capture_delay'] = 1
    p = eopayment.Payment('systempayv2', params)
    data = default_data.copy()
    data['vads_capture_delay'] = '3'
    transaction_id, f, form = p.request(**data)
    assert get_field(form, 'vads_capture_delay')['value'] == '3'

    # capture_date can be used for deferred_payment
    params = default_params.copy()
    params['vads_capture_delay'] = 1
    p = eopayment.Payment('systempayv2', params)
    data = default_data.copy()
    data['capture_date'] = datetime.now().date() + timedelta(days=4)
    transaction_id, f, form = p.request(**data)
    assert get_field(form, 'vads_capture_delay')['value'] == '4'


def test_manual_validation():
    params = {
        'secret_test': '1122334455667788',
        'vads_site_id': '12345678',
        'vads_ctx_mode': 'TEST',
    }
    data = {'amount': 15.24, 'orderid': '654321', 'first_name': 'John', 'last_name': 'Doe'}

    backend = eopayment.Payment('systempayv2', params)
    transaction_id, f, form = backend.request(**data.copy())
    assert get_field(form, 'vads_validation_mode')['value'] == ''

    data['manual_validation'] = True
    transaction_id, f, form = backend.request(**data.copy())
    assert get_field(form, 'vads_validation_mode')['value'] == '1'

    data['manual_validation'] = False
    transaction_id, f, form = backend.request(**data.copy())
    assert get_field(form, 'vads_validation_mode')['value'] == ''


FIXED_TRANSACTION_ID = '1234'


def test_transaction_id_request(backend):
    transaction_id, kind, form = backend.request(10.0, transaction_id=FIXED_TRANSACTION_ID)
    assert transaction_id == FIXED_TRANSACTION_ID
    found = None
    for field in form.fields:
        if field['name'] == 'vads_ext_info_eopayment_trans_id':
            found = field
            break
    assert found
    assert found['value'] == FIXED_TRANSACTION_ID


def test_transaction_id_response(backend, caplog):
    caplog.set_level(0)
    response = '''vads_amount=1000&vads_auth_mode=FULL&vads_auth_number=3fcdd2&vads_auth_result=00&vads_capture_delay=0&vads_card_brand=CB&vads_card_number=597010XXXXXX0018&vads_payment_certificate=4db13859ab429cb6b9bae7546952846efd190e3a&vads_ctx_mode=TEST&vads_currency=978&vads_effective_amount=1000&vads_effective_currency=978&vads_site_id=51438584&vads_trans_date=20201027212030&vads_trans_id=sDJJeQ&vads_trans_uuid=368ef4d0822448e3a2e7413c4e9f8be8&vads_validation_mode=0&vads_version=V2&vads_warranty_result=&vads_payment_src=EC&vads_cust_country=FR&vads_contrib=eopayment&vads_tid=001&vads_sequence_number=1&vads_contract_used=2334410&vads_trans_status=AUTHORISED&vads_expiry_month=6&vads_expiry_year=2021&vads_bank_label=Banque+de+d%C3%A9mo+et+de+l%27innovation&vads_bank_product=MCW&vads_pays_ip=FR&vads_presentation_date=20201027212031&vads_effective_creation_date=20201027212031&vads_operation_type=DEBIT&vads_result=00&vads_extra_result=&vads_card_country=FR&vads_language=fr&vads_brand_management=%7B%22userChoice%22%3Afalse%2C%22brandList%22%3A%22CB%7CMASTERCARD%22%2C%22brand%22%3A%22CB%22%7D&vads_action_mode=INTERACTIVE&vads_payment_config=SINGLE&vads_page_action=PAYMENT&vads_ext_info_eopayment_trans_id=1234&vads_threeds_enrolled=Y&vads_threeds_auth_type=CHALLENGE&vads_threeds_eci=02&vads_threeds_xid=bVpsTUhLSWpodnJjdXJVdE5rb0g%3D&vads_threeds_cavvAlgorithm=2&vads_threeds_status=Y&vads_threeds_sign_valid=1&vads_threeds_error_code=&vads_threeds_exit_status=10&vads_threeds_cavv=jG26AYSjvclBARFYSf%2FtXRmjGXM%3D&signature=fBGbFQPlUiyrL0yVgQzbhokMt6cqG24hOr%2BYsXKr/b8='''
    result = backend.response(response)
    assert result.signed
    assert result.transaction_id == FIXED_TRANSACTION_ID
